/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef COMP_H
#define COMP_H

#include "Transition.h"
#include "Formula.h"
#include "StateVar.h"
#include "ModeVar.h"
#include "CondVar.h"
#include "ObsVar.h"
#include "DependentVar.h"
#include "CommandVar.h"
#include "../caen/caen-objet.h"
#include "../system/purge.h"
#include <string>
#include <vector>

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

/**
 * \class Comp
 * \brief A component is a transition system.
 *
 * A component is made of state variables <StateVar>
 */

class Comp {
 public:
  string type;  /**< type of the component. */
  string name;  /**< name of the component. */
  unsigned int atoms_number;  /**< number of atoms in the component. */
  vector<CommandVar*>* cmd_in;   /**< commands in */
  vector<string>* str_in;        /**< commands values in */
  vector<CommandVar*>* cmd_out;  /**< commands out */
  vector<string>* str_out;
  vector<StateVar*>* state_variables;
  vector<Formula*>* constraints;
  vector<Clause*>* external_clauses;  /**< external clauses (compiled constraints) */
  vector<variable>* variables;        /**< continuous variables */
  vector<StateVar*>* others_state_variables;  /**< external discrete variables */
  vector<variable>* others_variables; /**< external continuous variables */
  vector<influence>* influences;  /**< continuous influences */
  vector<Transition*>* component_transitions; /**< transitions between modes */
  vector<Comp*>* imported_comps; /**< imported components */
  vector<pair<string, string> >* corres_import_comps;  /**< correspondances argument name in def / comp name in the instance */
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* atoms; /**< atoms references (used by model compiler).*/
  double nominal_cover;
  string var_name;   /**< variable name in the model */
  bool instanciated;
  bool generated;
  bool share_commanded;
  /* lists */
  listcaenInfl influListcaen;
  listcaenGVar gVarListcaen;
  listcaenInitVar initListcaen;
  listcaenExpr globCListcaen;
  listcaenExpr condCListcaen;
  int numCond;
  int numClauses;  /**< number of clauses in the component (for code generation purpose) */

 public:
  Comp();
  Comp(string na);
  Comp(string na, vector<StateVar*>* sta);
  Comp(string na, vector<CommandVar*> in, vector<CommandVar*>* out);
  ~Comp();

  void setStateVariables(vector<StateVar*>* sta);
  //add a state variable
  void addStateVar(StateVar* sta_var);
  /* add a continuous variables. */
  void addVariable(variable va);
  //return all current atoms.
  /*vector<Atom*>* getAllCurrentAtoms();*/
  //add an internal component.
  void addInternalComp(Comp* co);
  //compute nominal cover.
  void computeNominalCover();
  
  //set the vector of feasible states
  void setConstraints(vector<Formula*>* fea);
  //return the vector of feasible states
  vector<Formula*>* getConstraints();
  //add a Formula to feasible states
  void addFormulaToCS(Formula* frm);
  //add an influence.
  void addInfluence(influence inf);
  //set the vector of transitions
  void setCompTransitions(vector<Transition*>* trans);
  //add a Transition to the component's transitions
  void addTransitionToC(Transition* trans);

  //set the name of the component
  void setName(string na);

  /* tools for the parser */
  string findCorresDef (string name);
  string findCorresName (string defname);

  /**
   * Tools & debug.
   */
  void getTransitionsGuards();

  /**
   * debug & visu.
   */
  void showComp();

  /* inline methods */

/**
 * @return the set of state variables of the component
 */
inline vector<StateVar*>* Comp::getStateVariables() {
  return state_variables;
}
/**
 * @param na name of the variable.
 * @return a state variable, given its name.
 */
inline StateVar* Comp::getStateVariable(string na) {
  for (unsigned int i=0; i<state_variables->size(); i++)
    if ((*state_variables)[i]->name == na)
      return (*state_variables)[i];
  return NULL;
}
/**
 * @return all the atoms of the component.
 * \warning the method creates a vector of <Atom> as a result. 
 * this vector must be deleted after use for no memory leaks.
 */
inline vector<Atom*>* Comp::getAllAtoms() {
  vector<Atom*>* res = new vector<Atom*>();
  vector<Atom*>* temp = NULL;
  for (unsigned int i=0; i<state_variables->size(); i++) {
    temp = (*state_variables)[i]->getAtomLinks();
    for (unsigned int j=0; j<temp->size(); j++)
      res->push_back((*temp)[j]);
  }
  delete temp;
  return res;
}

/**
 * @return the set of transitions of the component.
 */
inline vector<Transition*>* Comp::getCompTransitions() {
  return component_transitions;
}
/**
 * @return the name of the component casted as a const char*
 */
inline const char* Comp::getName() {
  return name.c_str();
}
/**
 * @return the variable name of the component casted as a const char*
 */
inline const char* Comp::getVarName () {
  return var_name.c_str ();
}

};

#endif
