/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* State for trajectories and itms results. */

/* 
CHANGES : 

27/11/00 : creation.

*/

#ifndef STATE_H
#define STATE_H

#include "TrajEl.h"
#include "Comp.h"
#include "../system/config.h"
#ifdef USE_BDD
#include "bdd.h"
#endif

#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map>
#endif

#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif


/**
 * \struct eqstr
 * Equality operator for const char* for hashtables (hash_map).
 */

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

/**
* \class State
* \brief state of the physical system at a given time step.
* 
* A State is a set of instances at a given time step of some components' 
* modes from the model.
*/ 

class State {
  
 public:
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state;
  double rank; /* for ITMS internals only. */
#ifdef PROBAS
  double cover;
  double proba; /* p(O/s)*P(T)*p(st-1). */
  double prev_trans_prob; /* current probability of independent probabilities to that state. */
  double obs_cond_prob; /* p(O/s). */
  double static_prob; /* p(s). */
#endif
  unsigned int atoms_number; /* number of atoms in this state. */
#ifdef USE_BDD
  bdd* pb; /* state problem expressed as a BDD. */
  bdd* bdd_state; /* representation of the state as a bdd; */
  bdd* no_obs; /* represents state without obs. */
  bdd* obs; /* represents the state with observables. */
#endif
#ifdef PROBAS
  hash_map<int, double, hash<int> >* trans_probs; /* storage for transitions posterior probabilities. */
  multimap<double, int, greater_equal<double> >* sorted_trans_probs;
  hash_map<const char*, hash_map<int, double, hash<int> >*, hash<const char*>, eqstr>* temp_probs;
#endif
  hash_map<const char*, Comp*, hash<const char*>, eqstr>* comps;
  bool st_lock;
  bool observable; /* flag for state observability, i.e. if it contains observations or not. */

#ifdef RECONF
  vector<Atom*>* modifs;   /* false atoms modifications. */
#endif

  /* test & debug. */
  static int state_created;
  static int state_del;
  

  State();
  State(hash_map<const char*, Atom*, hash<const char*>, eqstr>* first_state);
  State(State* sta, bool locks);
  ~State();
  TrajEl* addElt(TrajEl* elt, bool lock);
  bool evalStateFormula(StateFormula* stfrm, State* sta);
  
  /**
   * methods for probabilities computation.
   */
#ifdef PROBAS
  double getTransitionPostProb(Transition* trans);
  void computePrevTransProb();
  double computePrevTransGuard();
  double computeNextTransGuard(); //useless.
#endif  

  /**
   * Debug & representation.
   */
  void showState();
  void showFullState();

/* inline methods */
/**
 * search an element in a state.
 * @param id the identity of the atom (e.g. v.mode=closed)
 * @return the corresponding element
 */
inline TrajEl* State::getElt(const char* id) {
  if (state->count(id) == 0)
    return NULL;
  else
    return (*state)[id];
}
/**
 * set the prevTrans to an elt. 
 * @param id name of the element.
 * @param trans Transition to set.
 */
inline void State::setPrevTrans(const char* id, Transition* trans) {
  (*state)[id]->prevTrans = trans;
}
/* set the prevTrans to an elt (should be in state !!!), and update transitions_cover. */
inline void State::setPrevTrans(TrajEl* elt, Transition* trans) {
  elt->prevTrans = trans;
}
/**
 * clear the state before it has to be deleted. 
 */
inline void State::clear() {
  if (state != NULL) {
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator j = state->begin(); j != state->end(); j++) {
      delete (*j).second;
      (*j).second = 0;
    }
  }
  delete state;
  state = 0;
}
};
#endif

