/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Element of a trajectory's state = atom + links. */

/*
- 18/11/00 : support init bug corrected.
 */

#ifndef TRAJEL_H
#define TRAJEL_H

#include "Transition.h"
#include "Clause.h"
#include "../system/purge.h"

/**
 * \class TrajEl
 * \brief A TrajEl is the minimal element in a <Trajectory> : a <State> is a set
 * of TrajEl, and a Trajectory is a set of State. 
 *
 * A TrajEl has a label: true, false, unknown. Each TrajEl is related to an Atom.
 * Atoms are static knowledge and serve as information support: when building trajectories,
 * they are instanciated as TrajEl. Note that TrajEl doesn't inherit Atom to not recopy information
 * in memory. (Another solution would be to create a static class for each atom).
 *
 * A TrajEl's label can be locked if its label shouldn't be changed again. They are two locking levels:
 * one is permanent, the other isn't (see lock for detail).
 */

class TrajEl { //minimal element of a trajectory.
 public:
  Atom* at; //atom, from a ModeVar.
  int prev_step; //timestep of the last modification of the label.
  Transition* prevTrans; //transition that support the atom in the trajectory.
  int prevTrans_time; //time the prevTrans has been triggered.
  map<int, Transition*>* nextTrans;
  map<int, Transition*>* nextTransNeg; //for support of negated atoms.
  string label; //ITMS label (true, false, unknown).
  Clause* support; //ITMS support.
  vector<TrajEl*>* supported; //other elements supported by this one.
  int propagation_number; //propagation number for fast dependencies evaluation.
  int lock; //0 is lock-true, 1 is lock-false, 2 is unknown (init value), 3 is permanent lock-true, 4 is permanent lock-false.
  double proba;
  /* debug. */
  static int trajel_created; /* number of elts created. */
  static int trajel_del; /* number of elts deleted. */

  TrajEl();
  TrajEl(TrajEl* el);
  TrajEl(Atom* a, Transition* tprev, map<int, Transition*>* tnext, string lab);
  ~TrajEl();
  void setAtom(Atom* a);
  bool eq(TrajEl& el);
  bool evalTrajEl();
  Atom* getAtomLink(string id); //get an atom from the state variable.
  void addToNextTrans(Transition* trans);
  void addToNextTransNeg(Transition* trans);
  void updateProba(double prob);
  void showAtom(bool ev);
  void showTrajEl(bool ev);

/* inline methods */

/**
 * @return the id of the elt (atom's id).
 */
inline const char* TrajEl::getId() {
  return at->getId();
}
/**
 * @return the atom's state variable.
 */
inline StateVar* TrajEl::getStateVar() {
  return at->getStateVar();
}
/**
 * @return the type of the state variable.
 */
inline string TrajEl::getVarType() {
  return at->getVarType();
}
/**
 * @return the label of the elt.
 */
inline string TrajEl::getLabel() {
  return label;
}
/**
 * set the label of the elt.
 * @param lab is the label to be set: true, false, unknown.
 */
inline void TrajEl::setLabel(string lab) {
  label = lab;
}

/**
 * @return the component the element belongs to.
 */
inline Comp* TrajEl::getVarComp () {
  return at->getVarComp ();
}

};

#endif
