/*
 * CHANGES:
 *
 *  05/06/02 : - modified unit clause scan => broader and slower search (for prime implicants generation).
 *  31/01/01 : - unitPropagate : move agenda from 'queue' to 'stack' for better results.
 *  18/12/00 : - added atoms post probabilities compilation from transitions
 *                 (compileTransitionsProb()).
 *  21/11/00 : - adding default clause (general one) to atoms during compilation.
 *  18/11/00 : - fixing bug on optimized getFirstUnit.
 */

#ifndef ITMS_H
#define ITMS_H

#define SUP_ITMS

/*#define DEBUG_ITMS*/

#include "../behav/State.h"
#include "../behav/Comp.h"
#include "../behav/Clause.h"
#include "../behav/Atom.h"
#include "../behav/TrajEl.h"
#include "../behav/StateVar.h"
#include "../behav/StateFormula.h"
#include "../behav/Formula.h"
#include "../behav/Transition.h"
#include "../behav/Trajectory.h"
#include "../system/purge.h"
#include "../system/config.h"
#include "mdlChk.h"
#include "pi_table.h"
#include <time.h>
#include <vector>
#include <map>
#include <queue>
#include "math.h"

#undef PROBAS
#undef ENTROPY

/**
 * \struct eqstr
 * Equality operator for const char* for hashtables (hash_map).
 */

#ifndef EQSTR_S
#define EQSTR_S
struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};
#endif

extern char appliName[];

/**
 * \class ItmsRes
 * \brief class for ITMS results.
 * A call to the ITMS returns a set of ItmsRes.
 * An ItmsRes is made for storing useful links that reduce 
 * the search by focusing the conflicts detection. The state resulting from the ITMS
 * work is in res.
 */

#ifndef ITMSRES_H
#define ITMSRES_H
class ItmsRes {
 public:
  
  State* res; /**< res is the valid new working state. */
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed; /**< already_changed is the set of elements that already have been modified by the ITMS during the current search. */
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg_tmp; /**< subset of already_changed of elements recently moved. */
  vector<TrajEl*>* deleted; /**< elements that where recently deleted. */
  queue<Clause*>* forced_cstrs; /**< clauses that are false in the State res. */

  /**
   * Empty constructor.
   */
  ItmsRes() {
    res = new State();
    forced_cstrs = new queue<Clause*>();
    already_changed = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Constructor to wrap an existing State.
   */
  ItmsRes(State* wkst) {  
    res = wkst;
    forced_cstrs = new queue<Clause*>();
    already_changed = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Constructor that makes a wrap around a copy of an existing State.
   */
  ItmsRes(State* wkst, queue<Clause*>* fc) {
    res = new State(wkst, 1);
    forced_cstrs = new queue<Clause*>(*fc);
    chg_tmp = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
    deleted = new vector<TrajEl*>();
  }
  /**
   * Partial destructor. Full destruction is possible by making use of clear() before
   * deleting the object.
   */
  ~ItmsRes() {
    delete already_changed;
    already_changed = 0;
    delete chg_tmp;
    chg_tmp = 0;
    delete deleted;
    deleted = 0;
    delete forced_cstrs;
    forced_cstrs = 0;
  }
  /**
   * delete the State res and its elements, as already_changed.
   */
  void clear() { 
    res->clear();
    delete res;
    res = 0;
  }
  /**
   * test is the current wrapped state is empty.
   * @result true if empty, false otherwise.
   */
  bool resEmpty() {
    return res->state->empty();
  }
};
#endif

/**
 * ITMS (incremental truth maintenance system) based on William/Nayak 1997.
 * 
 * \class itms
 * \brief Incremental Truth Maintenance System.
 *
 * The itms runs on an environment, which is made of a <State>, a <Trajectory> and the 
 * corresponding time-step. The environment is changed at each itms call.
 * A call to the itms is realized by using the <addProposition> function. Results are 
 * encapsulated into a set of <ItmsRes>.
 */

class itms : public mdlChk {

 public:
  vector<Formula*>* constraints; /**< set of constraints made of formulas. It is deleted after model has been compiled into clauses.*/
  vector<StateVar*>* state_variables; /**< in order to compile all clauses at once... (internal + external). */
  bool lock_comp; /**< boolean flag for component insertion, as unlocked or locked (for observations). */
  PiTable* pis;  /* prime implicants */

  itms();
  itms(hash_map<const char*, Atom*, hash<const char*>, eqstr>* set, vector<Formula*>* cstrs, hash_map<const char*, Comp*, hash<const char*>, eqstr >* comps);
  ~itms();

  /* get an elt. */
  TrajEl* getElt(const char* id, State* wkst, Trajectory* tr, int timestep, bool trans_idle);
  TrajEl* getElt(const char* id, State* wkst, Trajectory* tr, int timestep);
  /*************************
   * Knowledge Compilation *
   *************************/
  void initialize();
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* loadAtoms(vector<Clause*>* clauses);
  void idTransitions(Comp* co, int &curr_nu);
  void compileTransitionsProb(Atom* at);
  void compilePI();
  void compilePreInitAtoms (Atom* at, vector<StateFormula*>* cp);
  void compilePreAtoms (Atom* at, vector<StateFormula*>* cp);
  void clauseCompilation(); //compile the entire set of components.
  void clauseCompilation(vector<Formula*>* cstrs, int clause_num, Comp* co); //compile constraints' formulas into clauses.
  vector<Clause*>* compileFormula(Formula* frm, vector<Clause*>* res, Comp* co);
  vector<Clause*>* compileStateFormula(StateFormula* stfrm, bool non, Comp* co);
  vector<Clause*>* mergeClauses(vector<Clause*>* res1, vector<Clause*>* res2, vector<Clause*>* res, Comp* co);
  bool intersectClauses(vector<Clause*>* cls1, vector<Clause*>* cls2);
  bool intersectClauses(Atom* at, vector<Atom*>* vec);
  void distribClauses(Clause* res, StateFormula* stfrm);
  void rtInit ();
#ifdef PROBAS
  void initBddPb(State* wkst);
#endif

  /*********************************
   * Truth Maintenance over States *
   *********************************/
#ifdef SUP_ITMS
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* copyState(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, State* wkst);
  int maxPropagationNumber(Clause* cla, TrajEl* p, State* wkst, Trajectory* tr, int timestep); //compute the max of the propagation_number of a clause.
  bool isResupportable(TrajEl* p, Clause* cla, bool non, State* wkst, Trajectory* tr, int timestep); //test if resupporting is possible.
#endif
  bool inVector(Atom* at, vector<Atom*>* vec); //test if at is in vec.
  bool inVector(const char* id, vector<const char*>* vec);
  bool inVector(TrajEl* at, vector<TrajEl*>* vec);
  bool inVector(Transition* tr, vector<Transition*>* vec);
  bool inVectorStateVar(StateVar* st, vector<StateVar*>* vec); //test if a state variable is in a vec.
  bool inVectorClause(Clause* cl, vector<Clause*>* vec);
  bool inMultimap(Atom* at, multimap<bool, Atom*, eqbool>* mul);
  hash_map<const char*, StateVar*, hash<const char*>, eqstr>* getStateVars(vector<TrajEl*>* vec); //return the set of statevars from the set of elts.
#ifdef SUP_ITMS
  void setSupport(Clause* cla, TrajEl* at, State* wkst, Trajectory* tr, int timestep);
  multimap<bool, TrajEl*, eqbool>* getNotUnknown(Clause* cla, State* wkst, Trajectory* tr, int timestep);
  TrajEl* getUnknown(Clause* cla, State* wkst, Trajectory* tr, int timestep);
  string evalAtomOnLabels(multimap<bool, Atom*, eqbool>::iterator it, State* wkst, Trajectory* tr, int timestep);
  bool testClause(Clause* cla, State* wkst, Trajectory* tr, int timestep);
  TrajEl* unitClause(Clause* cla, State* wkst, Trajectory* tr, int timestep);
  Clause* getFirstUnitClause(Atom* at, State* wkst, Trajectory* tr, int timestep); //return the first unit clause to be found in at's dependencies.
  Clause* getFirstUnitClause(State* wkst, Trajectory* tr, int timestep); //return the first unit clause from the entire set of clauses.
  Clause* getFirstUnitClause(vector<TrajEl*>* deleted, State* wkst, Trajectory* tr, int timestep); //more efficient...
  void resupportDependence(TrajEl* at, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* resc, State* wkst, Trajectory* tr, int timestep); //resupport labels.
  void checkForFocusedInternalConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, int timestep);
  void checkForFocusedInternalConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, vector<Clause*>* forced_cstrs, Trajectory* tr, int timestep);
  void checkForFocusedInternalConflictsCBFS(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, vector<Clause*>* forced_cstrs, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* st);
  void checkForFocusedConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, int timestep);
  Clause* checkForInternalConflicts(TrajEl* at, State* wkst, Trajectory* tr, int timestep);
  void checkForConflicts(vector<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, int timestep);
  void propagateClause(Clause* cla, ItmsRes* curr_res, Trajectory* tr, int tiemstep);
  ItmsRes* propagateThroughConflict(Clause* cla, ItmsRes* curr_res, stack<ItmsRes*>*, bool nom, double rank, Trajectory* tr, int timestep); 
  void unitPropagate(ItmsRes* curr_st, vector<ItmsRes*>* res, bool nom, double rank, Trajectory* tr, int timestep, unsigned int num_res);
  vector<ItmsRes*>* addPropositions(multimap<bool, TrajEl*, eqbool>* add_ats, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed, State* wkst, vector<ItmsRes*>* res, bool nom, double rank, double current_rank, Trajectory* tr, int timestep, bool comp_lock);
  vector<ItmsRes*>* testImplicants(multimap<bool, TrajEl*, eqbool>* add_ats, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed, State* wkst, vector<ItmsRes*>* res, bool nom, double rank, double current_rank, Trajectory* tr, int timestep, bool comp_lock, unsigned int pos);
  vector<TrajEl*>* deleteClause(TrajEl* del_at, vector<TrajEl*>* res, State* wkst, Trajectory* tr, int timestep);
  vector<TrajEl*>* traceSupport(TrajEl* p, vector<TrajEl*>* res, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, bool neg, bool flag); //trace the support of an atom.
  vector<TrajEl*>* traceTransitionsSupport(TrajEl* p, vector<TrajEl*>* res, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, bool neg, bool flag);
  vector<Atom*>* predictedAtoms(Atom* p, vector<Atom*>* res); //get the atoms that are predicted by an atom.
  double probaCondObsAtom(Atom* at, vector<Atom*>* obs); //return an estimation of the conditional probability of the atom given a set of observations.

  /****************************
   * Probabilities management *
   ****************************/
#ifdef ENTROPY
  double atomProba(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double entropy(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double entropy(StateVar* st, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  double condProbTrans(Transition* trans, TrajEl* elt, State* wkst, double atom_prob, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  void initStateProbas(State* wkst);
  void sortPostProbs(State* wkst);
#endif

  /*******************************************************
   * Trajectories management and consistency maintenance *
   *******************************************************/
  bool evalStateFormula(StateFormula* stfrm, Trajectory* tr, int timestep);
#endif /* SUP_ITMS */

  int readPI(PiTable* pit);

  /*************
   * debugging *
   *************/
  void showClause(Clause* cl);
  void viewAtomVector(vector<Atom*>* vec);
  void viewAtomVectorId(vector<Atom*>* vec);
  void viewClauseVector(vector<Clause*>* vec);
  void viewMultimapAtom(multimap<string, Atom*, eqstring>* mlt_map);
  void viewHashMapAtom(hash_map<const char*, Atom*, hash<const char*>, eqstr>* set);
  void viewTrajectoryConflicts(hash_map<int, vector<Clause*>*, hash<int> >* confs);

/* inline methods */


};

#endif
