/*********************************************************************/
/*                                                                   */
/*                          --- optim.c ---                          */
/*                                                                   */
/* fct: module des procedures et fonctions utilisees par toutes les  */
/*      methodes d'optimisation non-lineaire.                        */
/*                                                                   */
/* aut: Francois Dufour (LAAS)                                       */
/*                                                                   */
/* ver: 22 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "optim.h"

/*********************************************************************/
/*                                                                   */
/*                         --- Optim_ran0 ---                        */
/*                                                                   */
/* fct: renvoie un nombre aleatoire compris entre 0 et 1.            */
/*                                                                   */
/* car: fonction decrite dans "Numerical recipes in C", page 205.    */
/*                                                                   */
/* ver: 17 juin 1993                                                 */
/*                                                                   */
/*********************************************************************/

double Optim_ran0
(
int *idum   /* ptr d'un nb entier quelconque: negatif pour initialiser la fonction */
)
{
   static double y, maxran, v[98];
   double dum;
   static int iff = 0;
   int j;
   unsigned i,k;

   if (*idum < 0 || iff == 0)
      {
      iff = 1;
      i = 2;
      do {
         k = i;
         i <<= 1;
         } while (i);

      maxran = k;
      srandom(*idum);
      *idum = 1;
      for (j=1; j<=97; j++) dum = random();
      for (j=1; j<=97; j++) v[j] = random();
      y = random();
      }
   j = 1 + 97.0*y/maxran;
   if (j > 97 || j < 1) 
      printf("\n!!! erreur dans Optim_ran0 !!!\n");
   y = v[j];
   v[j] = random();
   return (y/maxran);
}


/*********************************************************************/
/*                                                                   */
/*                      --- Optim_tri_index ---                      */
/*                                                                   */
/* fct: tri des indices d'un vecteur suivant les valeurs ascendantes */
/*      de ces composantes.                                          */
/*                                                                   */
/* car: - fonction decrite dans "Numerical recipes in C", page 248.  */
/*      - les indices varient de 0 a n-1.                            */
/*                                                                   */
/* ver: 22 fevrier 1994                                              */
/*                                                                   */
/*********************************************************************/

void Optim_tri_index
(
int     n,         /* dimension du vecteur */
double *vecteur,   /* ptr du vecteur a trier */
int    *index      /* ptr du vecteur des indices */
)
{
   int i, l, j , ir, index_temp, cor;
   double valeur;

/*** initialisation du vecteur des indices ***/

   for (j = 1; j <= n; j ++)
      index[j-1] = j-1;

/*** tri des indices ***/

   if (n > 1)
      {
      l = (n >> 1) + 1;
      ir = n;
      for (;;)
         {
         if (l > 1)
            valeur = vecteur[(index_temp = index[--l-1])];
         else
            {
            valeur = vecteur[(index_temp = index[ir-1])];
            index[ir-1] = index[0];
            if (--ir == 1)
               {
               index[1-1] = index_temp;
               return;
               }
            }

         i = l;
         j = l << 1;
         while (j <= ir)
            {
            if ((j < ir) && (vecteur[index[j-1]] < vecteur[index[j]]))
               j++;
            if (valeur < vecteur[index[j-1]])
               {
               index[i-1] = index[j-1];
               j += (i = j);
               }
            else
               j = ir + 1;
            }
         index[i-1] = index_temp;
         }
      }
}


/*********************************************************************/
/*                                                                   */
/*                     --- Optim_longvec_norm ---                    */
/*                                                                   */
/* fct: calcul de la longueur normalise (de 0.0 a 1.0) d'un vecteur. */
/*                                                                   */
/* ver: 2 fevrier 1994                                               */
/*                                                                   */
/*********************************************************************/

double Optim_longvec_norm
(
int     dimX,     /* dimension du vecteur            */
double *Xini,     /* ptr du point initial du vecteur */
double *Xfin,     /* ptr du point final du vecteur   */
double *delta_X   /* ptr des largeurs des domaines   */
)
{
   int i;
   double somme, temp;

   somme = 0.0;
   for (i = 0; i < dimX; i++)
      if (delta_X[i] > 0.0)
         {
         temp = (Xfin[i] - Xini[i])/delta_X[i];
         somme += temp*temp;
         }
   return (sqrt(somme));
}


/*********************************************************************/
/*                                                                   */
/*                         --- iveridom ---                          */
/*                                                                   */
/* fct: verifie le domaine d'admissibilite des variables             */
/*      d'optimisations.                                             */
/*                                                                   */
/* car: retourne 1 si admissible, sinon 0.                           */
/*                                                                   */
/* ver: 20 juillet 1993                                              */
/*                                                                   */
/*********************************************************************/

boolean iveridom
(
int    *n,      /* ptr du nombre de parametres de commande   */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax    /* ptr des bornes maximales du domaine des X */
)
{
   int i;
   boolean dom_admissible;

   dom_admissible = TRUE;
   for (i = 0; i < *n; i++)
      if ((X[i] < Xmin[i]) || (X[i] > Xmax[i]))
         dom_admissible = FALSE;

   return (dom_admissible);
}


/*********************************************************************/
/*                                                                   */
/*                          --- iveridom_ ---                        */
/*                                                                   */
/* fct: verifie le domaine d'admissibilite des variables d'optimisa- */
/*      tions.                                                       */
/*                                                                   */
/* uti: exemple d'appel par un programme Fortran:                    */
/*         itest = iveridom(n, X, Xmin, Xmax)                        */
/*                                                                   */
/* car: retourne 1 si admissible, sinon 0.                           */
/*                                                                   */
/* ver: 20 juillet 1993                                              */
/*                                                                   */
/*********************************************************************/

boolean iveridom_
(
int    *n,      /* ptr du nombre de parametres de commande */
double *X,      /* ptr du vecteur des parametres de commande */
double *Xmin,   /* ptr des bornes minimales du domaine des X */
double *Xmax    /* ptr des bornes maximales du domaine des X */
)
{
   return iveridom(n, X, Xmin, Xmax);
}


/*********************************************************************/
/*                                                                   */
/*                           --- drandom_ ---                        */
/*                                                                   */
/* fct: retourne une valeur aleatoire comprise entre 0 et 1.         */
/*                                                                   */
/* uti: exemple d'appel par un programme Fortran:                    */
/*         drand = drandom(1)                                        */
/*                                                                   */
/* ver: 20 juillet 1993                                              */
/*                                                                   */
/*********************************************************************/

double drandom_
(
int *idum   /* ptr d'un nb entier quelconque: negatif pour initialiser la fonction */
)
{
   return (Optim_ran0(idum));
}

/*** FIN ***/
