/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Diagnoser.h"

/* #define DEBUG_DIAG
   #define DEBUG */

/**
 * \class Diagnoser
 */

hash_map<int, Transition*, hash<int> >* Diagnoser::redux_diag = new hash_map<int, Transition*, hash<int> > ();

/* constructors. */
Diagnoser::Diagnoser() : itms() {

}
/**
 * Insert incoming observations into the current state of a tracked trajectory.
 */
void Diagnoser::setObservation(Atom* obs, State* sta) {
  StateVar* stavar = obs->getStateVar();
  /* set the corresponding element to 'true'. */
  (*sta->state)[obs->getId()]->setLabel("true");
  /* set the others 'twin' elements to 'false'. */
  for (unsigned int i=0; i<stavar->getAtomLinks()->size(); i++) {
    (*sta->state)[(*stavar->getAtomLinks())[i]->getId()]->setLabel("false");
  }
}

/****************************
 * Trajectories management. *
 ****************************/

/**
 * Locking system => there are two kind of locks:
 *      - non-permanent locks for internal algorithms:
 *              * false: lock=1;
 *              * true:  lock=0
 *      - permanent locks for observations and exogeneous commands:
 *              * false: lock=4
 *              * true:  lock=3
 *      - otherwise lock=2
 */

/*lock elts from the state that has the same state variable as the argument. */
void Diagnoser::lockTrajEls(TrajEl* elt, State* sta) {
  /* lock the argument elt. */
  if (elt->lock == 2) {
    for (unsigned int i=0; i<elt->getStateVar()->getAtomLinks()->size(); i++) {
      if ((*elt->getStateVar()->getAtomLinks())[i] != elt->at) {
	(*sta->state)[(*elt->getStateVar()->getAtomLinks())[i]->getId()]->lock = 1;
      }
    }
    elt->lock = 0;
  }
  /* mark the state as locked.*/
  sta->st_lock = true;
}

/* permanent lock for observables and exogeneous commands. */
void Diagnoser::lockTrajElsPerm(TrajEl* elt, State* sta) {
  /* lock the argument elt. */
  if (elt->lock == 2) {
    for (unsigned int i=0; i<elt->getStateVar()->getAtomLinks()->size(); i++) {
      if ((*elt->getStateVar()->getAtomLinks())[i] != elt->at) {
	(*sta->state)[(*elt->getStateVar()->getAtomLinks())[i]->getId()]->lock = 4;
      }
    }
    elt->lock = 3;
  }
  /*sta->rank = 0;*/ /* ???? desactivated 5/12/01 */
  sta->st_lock = true; /* not sure */
}

/* neg: 1 the atom to be locked is negated */
void Diagnoser::lockTrajElsPerm (TrajEl* elt, State* sta, bool neg) {
  if (! neg)
    lockTrajElsPerm (elt, sta);
  else lockTrajElsNegPerm (elt, sta);
}

void Diagnoser::lockTrajElsNegPerm (TrajEl* elt, State* sta) {
  elt->lock = 4;
  sta->st_lock = true;
}

/* unlock a state's locked elt except for its obsvar... */
void Diagnoser::unlockState(State* st) {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = st->state->begin(); it != st->state->end(); it++) {
    if ((*it).second->lock == 0 | (*it).second->lock == 1)
      (*it).second->lock = 2;
  }
  st->st_lock = false;
}

/**
* Unlock locked states in a trajectory (a solution).
* Could be improved to directly focus on modifed states...
*/
void Diagnoser::unlockTrajectory(Trajectory* tr) {
  for (hash_map<int, State*, hash<int> >::iterator it = tr->traj->begin(); it != tr->traj->end(); it++) {
    (*it).second->rank = 0;
    if ((*it).second->st_lock)
      unlockState((*it).second);
  }
  tr->rank = 0;
}

/******************************************** 
 *  Methods for the diagnosis engine. * 
 ********************************************/

/**
 * construct sols (trajectories) for transitions from the likely transitions set.
 * As it is possible for a transition to be in more than a solution for a given rank,
 * a set of trajectories is return.
 * @param curr_trans is the transition to be added.
 * @param tr is the trajectory to which the transition has to be added.
 * @param time_step is the current time step the system is working on.
 * @param rank is the current rank of the search.
 * @return a set of new trajectories.
 */
multimap<double, Trajectory*, greater<double> >* Diagnoser::constructSuccessors(likelyTransition* lktrans, Trajectory* tr, int time_step, multimap<double, Trajectory*, greater<double> >* res, double rank) {
  vector<ItmsRes*>* tracked_states_pre = new vector<ItmsRes*>();
  vector<ItmsRes*>* tracked_states_post = new vector<ItmsRes*>();
  vector<ItmsRes*>* add_cl_res; 
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* itms_res;
  TrajEl* curr_elt_res;
  TrajEl* tmpel;
  Transition* curr_trans = lktrans->trans;
  int start_step = lktrans->start_step;
  multimap<bool, Atom*, eqbool>* trans_pre = curr_trans->pre->getAllSignedAtoms();
  multimap<bool, Atom*, eqbool>* trans_post = curr_trans->post->getAllSignedAtoms();
  State* previous_state;
  State* current_state;
  Trajectory* n_traj;
  
#ifdef DEBUG_DIAG
    std::cerr << "construct begin :\n";
    std::cerr << "timestep : " << time_step << std::endl;
    std::cerr << "startstep : " << start_step << std::endl;
    curr_trans->showTransition();
#endif
  
    /* copy the states. */
    previous_state = new State(tr->getState(start_step), 1);
    current_state = new State(tr->getState(time_step), 1);
    double prev_rank = tr->getStateRank(tr->getState(start_step));
    double curr_rank = tr->getStateRank(tr->getState(time_step));
    multimap<bool, TrajEl*, eqbool>* trans_pre_props = new multimap<bool, TrajEl*, eqbool>();
    
    /* add the required component (of the transition) if they are not present in both states. */
    if ((*previous_state->comps)[curr_trans->modevar->getComp()->getName()] == NULL) {
      tr->addComp(curr_trans->modevar->getComp(), previous_state, start_step, 1, 0);
#ifdef USE_BDD
      addToStateBdd(previous_state, curr_trans->modevar->getComp()->getName());
#endif
    }
    if ((*current_state->comps)[curr_trans->modevar->getComp()->getName()] == NULL) {
      tr->addComp(curr_trans->modevar->getComp(), current_state, time_step, 0, 0);
#ifdef USE_BDD
      addToStateBdd(current_state, curr_trans->modevar->getComp()->getName());
#endif
    }
    
    /* updating the previous state using the 'pre' from the transition. */
    for (multimap<bool, Atom*, eqbool>::iterator i=trans_pre->begin(); i != trans_pre->end(); i++) {
      tmpel = (*previous_state->state)[(*i).second->getId()];
      if ((*i).first && tmpel->getLabel() != "true") {
	trans_pre_props->insert(pair<bool, TrajEl*>(1, tmpel));
	lockTrajEls(tmpel, previous_state);
	tmpel->addToNextTrans(curr_trans);
      }
      else if (!(*i).first && tmpel->getLabel() != "false") {
	trans_pre_props->insert(pair<bool, TrajEl*>(0, tmpel));
	tmpel->addToNextTransNeg(curr_trans);
      }
      else if ((*i).first) {
	lockTrajEls(tmpel, previous_state);
	tmpel->addToNextTrans(curr_trans);
      }
      else if (!(*i).first)
	tmpel->addToNextTransNeg(curr_trans);
    }
    /* adding the propositions to previous state. */
    add_cl_res = addPropositions(trans_pre_props, NULL, previous_state, NULL, 0, rank, tr->rank, tr, start_step, 0);
    if (add_cl_res->empty()) {
#ifdef DEBUG_DIAG
      std::cerr << ">>>>> Transition aborted : \n"; 
      curr_trans->showTransition();
#endif
      /* previous state already deleted by the itms... */
      current_state->clear();
      delete current_state;
      current_state = 0;
      delete trans_pre; 
      trans_pre = 0;
      delete trans_post;
      trans_post = 0;
      delete tracked_states_pre;
      tracked_states_pre = 0;
      delete tracked_states_post;
      tracked_states_post = 0;
      delete trans_pre_props;
      trans_pre_props = 0;
      delete add_cl_res;
      add_cl_res = 0;
      return res;
    }
    for (unsigned int k=0; k<add_cl_res->size(); k++) {
      tracked_states_pre->push_back((*add_cl_res)[k]);
      previous_state = (*add_cl_res)[k]->res;
      itms_res = (*add_cl_res)[k]->already_changed;
      for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator itms_res_it = itms_res->begin(); 
	   itms_res_it != itms_res->end(); itms_res_it++) {
	curr_elt_res = (*previous_state->state)[(*itms_res_it).first];
	if (curr_elt_res->label != "false" ) {
	  lockTrajEls(curr_elt_res, previous_state);
	}
      }
    }
    delete add_cl_res;
    add_cl_res = 0;
    delete trans_pre_props;
    trans_pre_props = 0;
    
    multimap<bool, TrajEl*, eqbool>* trans_post_props = new multimap<bool, TrajEl*, eqbool>();
    for (multimap<bool, Atom*, eqbool>::iterator m=trans_post->begin(); m != trans_post->end(); m++) {
      tmpel = (*current_state->state)[(*m).second->getId()];
      if ((*m).first && tmpel->getLabel() != "true")
	trans_post_props->insert(pair<bool, TrajEl*>(1, tmpel));
      else if (!(*m).first && tmpel->getLabel() != "false")
	trans_post_props->insert(pair<bool, TrajEl*>(0, tmpel));
      current_state->setPrevTrans((*m).second->getId(), curr_trans);
      lockTrajEls(tmpel, current_state);
      /* set the prevTrans of the elt from the same state var to NULL. */
      for (unsigned int z=0; z<(*m).second->getStateVar()->getAtomLinks()->size(); z++) {
	if ((*current_state->state)[(*(*m).second->getStateVar()->getAtomLinks())[z]->getId()] != tmpel) {
	  (*current_state->state)[(*(*m).second->getStateVar()->getAtomLinks())[z]->getId()]->prevTrans = NULL;
	}
      }
    }
    /* adding the propositions to current state. */
    add_cl_res = addPropositions(trans_post_props, NULL, current_state, NULL, 0, rank, tr->rank, tr, time_step, 0);
    for (unsigned int o=0; o<add_cl_res->size(); o++) {
      tracked_states_post->push_back((*add_cl_res)[o]);
      current_state = (*add_cl_res)[o]->res;
      itms_res = (*add_cl_res)[o]->already_changed;
      for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator itms_res_it2 = itms_res->begin(); 
	   itms_res_it2 != itms_res->end(); itms_res_it2++) {
	curr_elt_res = (*current_state->state)[(*itms_res_it2).first];
	if (curr_elt_res->label != "false") {
	  lockTrajEls(curr_elt_res, current_state);
	}
      }
    }
    delete add_cl_res;
    add_cl_res = 0;
    delete trans_post_props;
    trans_post_props = 0;
    
#ifdef DEBUG_DIAG
    std::cerr << "time_step :" << time_step << std::endl;
    std::cerr << "trans added :";
    curr_trans->showTransition();
#endif
    
    /* create the new trajectories... */
    for (unsigned int p=0; p<tracked_states_pre->size(); p++) {
    
#ifdef DEBUG_DIAG
      std::cerr << "tracked_states_pre :\n";
      viewHashMapTrajElLow((*tracked_states_pre)[p]->res->state);
      std::cerr << std::endl;
#endif

    for (unsigned int q=0; q<tracked_states_post->size(); q++) {

#ifdef DEBUG_DIAG
      std::cerr << "trackec_states_post :\n";
      viewHashMapTrajElLow((*tracked_states_post)[q]->res->state);
      std::cerr << std::endl;
#endif
      n_traj = new Trajectory(tr, tr->getCurrentTime()+1);
      n_traj->rank = n_traj->rank - prev_rank - curr_rank;  /* update trajectory rank. */
      n_traj->addCopyState(start_step, (*tracked_states_pre)[p]->res, 1);
      n_traj->addCopyState(time_step, (*tracked_states_post)[q]->res, 1);
      if (n_traj->trans_diag == NULL)
	n_traj->trans_diag = new multimap<int, Transition*> ();
      n_traj->trans_diag->insert(pair<int, Transition*>(time_step, curr_trans));
      res->insert(pair<double, Trajectory*>(n_traj->rank, n_traj));
      if (p == tracked_states_pre->size() - 1) {  /* remove some components if required. */
	(*tracked_states_post)[q]->clear();
	delete (*tracked_states_post)[q];
	(*tracked_states_post)[q] = 0;
      }
    }
    (*tracked_states_pre)[p]->clear();
    delete (*tracked_states_pre)[p];
    (*tracked_states_pre)[p] = 0;
  }
  delete lktrans;
  lktrans = 0;
  delete trans_pre;
  trans_pre = 0;
  delete trans_post;
  trans_post = 0;
  delete tracked_states_pre;
  tracked_states_pre = 0;
  delete tracked_states_post;
  tracked_states_post = 0;

#ifdef DEBUG_DIAG
    std::cerr << "construct size : " << res->size() << std::endl;
    std::cerr << std::endl;
    //compareTrajs(res);
#endif

  return res;
}

/** 
 * Compute the successors of a sol. The search is focused with a set of conflicts.
 * 
 * @param soln is the detected conflict.
 * @param tr is the currently studied trajectory with conflict.
 * @param time_step is the current time step the system is working on.
 * @param rank is the current rank limit of the search.
 * @param trajs is the set of trajectories results as the method is a recursive one.
 * @param post defines if the search is made from pre to post, or from post to pre.
 * @return a set of consistent trajectories between time_step and time_step-1 (post search) or time_step+1 (pre search).
 */
multimap<double, Trajectory*, greater<double> >* Diagnoser::immediateSuccessors(hash_map<const char*, disConflict*, hash<const char*>, eqstr>* soln, Trajectory* tr, int time_step, double rank, multimap<double, Trajectory*, greater<double> >* trajs, bool post) {
  State* state;
  State* pre_state;
  double trans_rank = 0;
  double current_rank = tr->rank;
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* conf_pre_next = 0;// = new hash_map<const char*, disConflict*, hash<const char*>, eqstr>();
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* conf_post_prev = 0;// = new hash_map<const char*, disConflict*, hash<const char*>, eqstr>();
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* conf_post_next = 0;// = new hash_map<const char*, disConflict*, hash<const char*>, eqstr>();
  multimap<double, likelyTransition*, greater<double> >* likely_trans = new multimap<double, likelyTransition*, greater<double> >();
  multimap<double, Trajectory*, greater<double> >* n_trajectories = new multimap<double, Trajectory*, greater<double> >();
  multimap<bool, Atom*, eqbool>* all_tmp;
  StateVar* curr_stv;
  vector<StateVar*>* stas = new vector<StateVar*>(); //statevar already processed. Could be in GetConf...
  TrajEl* curr_val; //current value.
  TrajEl* curr_at = NULL; //end value.
  TrajEl* tmpel;
  Transition* curr_trans; //currently evaluated transition.
  bool lock_flag = false;
  vector<Transition*>* at_trans;
  bool break_flag = false;
  bool delete_flag = true;
  int start_step; /* starting step for long id transitions. */
  int next_step; /* state to probe into the 'future' in case it is needed for consistency. */
  bool sign;
  ModeVar* cmd_modevar = NULL; /* in case the conflict is a command, only perform the search on a single mode var.*/
  /*Trajectory* routed_traj;*/
  likelyTransition* liketrans;

  /* get the most likely transitions to be taken. */
  for (hash_map<const char*, disConflict*, hash<const char*>, eqstr>::iterator it_soln = soln->begin(); 
       it_soln != soln->end(); it_soln++) {
    curr_val = (*it_soln).second->conf;
    start_step = (*it_soln).second->conf_time;
    sign = (*it_soln).second->sign;
    curr_val->nextTrans->clear();
    if (post && curr_val->prevTrans != NULL) {
#ifdef HANDLE_TCM
      /* in case it is a multiple commands transition, update other states. 
	 i.e. : generate 'itms-level' conflicts on other linked values over the past time.
      */
      if (curr_val->prevTrans->post == NULL) {
	for (hash_map<int, StateFormula*, hash<int> >::iterator pp_it = curr_val->prevTrans->st_post->begin(); 
	     pp_it != curr_val->prevTrans->st_post->end(); pp_it++) {
	  if ((curr_val->prevTrans_time + (*pp_it).first) <= time_step) {
	    for (multimap<bool, Atom*, eqbool>::iterator pp = (*pp_it).second->atoms->begin(); pp != (*pp_it).second->atoms->end(); pp++) {
	      temp_elt = tr->getElt(curr_val->prevTrans_time + (*pp_it).first, (*pp).second->getId());
	      if (temp_elt->lock == 2) temp_elt->label = "false";
	    }
	  }
	}
      }
#endif
      if (curr_val->lock == 2 && curr_val->prevTrans->post != NULL ) {
	curr_val->prevTrans = NULL;  /* clear the prevTrans link of non locked atoms. */
      }
    }
    curr_stv = curr_val->getStateVar();
    if (!inVectorStateVar(curr_stv, stas)) {
      stas->push_back(curr_stv);
      state = tr->getState(time_step);
      pre_state = tr->getState(start_step);
      for (unsigned int k=0; k<curr_stv->getAtomLinks()->size(); k++) {
	if (sign && post)
	  curr_at = state->getElt((*curr_stv->getAtomLinks())[k]->getId());
	else if (sign && !post)
	  curr_at = pre_state->getElt((*curr_stv->getAtomLinks())[k]->getId());
	else if (!sign && post) {
	  curr_at = state->getElt(curr_val->getId());
	  break;
	  /*k = curr_stv->getAtomLinks()->size();*/
	}
	else if (!sign && !post) {
	  curr_at = pre_state->getElt(curr_val->getId());
	  break;
	  /*k = curr_stv->getAtomLinks()->size();*/
	}

	if (! sign || curr_at != curr_val) {
	  if (post) at_trans = curr_at->at->getRefTransPost();
	  else at_trans = curr_at->at->getRefTransPre();
#ifdef DEBUG_DIAG
	    std::cerr << "curr_at : " << curr_at->getId() << std::endl;
	    std::cerr << "ref trans size : " << at_trans->size() << std::endl;
#endif
	    for (unsigned int m=0; m<at_trans->size(); m++) {
	      lock_flag = false;
	      
	      /* begin the tests to choose the correct transitions. */
	      if (curr_val->getVarType() == "Command" 
		  && curr_val->at->value != "none" 
		  && curr_at->at->value != "none")  /* speed-up over commands => no change from a command to something else than none. */
		lock_flag = true;
	      
#ifdef ENTROPY
	      /*if (!state->trans_probs->empty() & state->getTransitionPostProb(curr_trans) == 0)
		lock_flag = true;*/
#endif
	      
	      /* one modevar/command variable during the search. */
	      if (!lock_flag && cmd_modevar != NULL && cmd_modevar != (*at_trans)[m]->modevar)
		lock_flag = true;
	      curr_trans = (*at_trans)[m];  /* looking at a trans. */
#ifdef DEBUG_DIAG
	    curr_trans->showTransition();
#endif    
	    if (!lock_flag && post && curr_at->prevTrans != NULL)  /* check if the trans is not the actual one. */
	      if (curr_trans == curr_at->prevTrans) {
#ifdef DEBUG_DIAG
		cerr << "trans is the actual one !\n";
#endif
		lock_flag = true;
	      }
	    if (!lock_flag && !post && ((*curr_at->nextTrans)[curr_trans->trans_id] != NULL 
		|| (*curr_at->nextTransNeg)[curr_trans->trans_id] != NULL)) {
#ifdef DEBUG_DIAG
		cerr << "curr_trans already in nextTrans !\n";
		curr_trans->showTransition();
#endif
	      lock_flag = true;
	    }
	    if (!lock_flag) {
	      for (multimap<double, likelyTransition*, greater<double> >::iterator lit = likely_trans->begin();  
		   lit != likely_trans->end(); lit++) { /* test if the transition has already been used in this session. */
		if (curr_trans->trans_id == (*lit).second->trans->trans_id) {
		  lock_flag = true;
		  break;
		}
	      }
	    }
	    if (!lock_flag) {  /* must check at timestep-1 if elts corresponding to pre of the transition are not locked. */
	      all_tmp = curr_trans->pre->getAllSignedAtoms();
	      for (multimap<bool, Atom*, eqbool>::iterator n = all_tmp->begin(); n != all_tmp->end(); n++) {
#ifdef DEBUG_DIAG
		  cerr << (*n).second->getId() << endl;
		  cerr << "timestep : " << time_step << endl;
		  cerr << "start_step : " << start_step << endl;
#endif
		  tmpel = getElt ((*n).second->getId (), pre_state, tr, start_step+1, 1);
		if ((*n).first && (tmpel->lock == 1 || tmpel->lock == 4)) {
#ifdef DEBUG_DIAG
		  cerr << (*n).second->getId() << " is locked !!\n";
#endif
		  lock_flag = true;
		  break;
		}
		else if (!(*n).first && (tmpel->lock == 0 || tmpel->lock == 3)) {
#ifdef DEBUG_DIAG		  
		  cerr << (*n).second->getId() << " is locked !!\n";
#endif		  
		  lock_flag = true;
		  break;
		}
		if ((*tmpel->nextTrans)[curr_trans->trans_id] != NULL) {  /* check if an elt of the transition already has support for it. */
#ifdef DEBUG_DIAG 
		  cerr << "already supported by : " << (*n).second->getId() << endl;
#endif		  
		  lock_flag = true;
		  break;
		}
	      }
	      delete all_tmp;
	      all_tmp = 0;
	      }
	      if (curr_trans->post == NULL) lock_flag = true; //patch.
	      if (!lock_flag) {             /* must check at timestep if elts in post of the transition are not locked. */
		all_tmp = curr_trans->post->getAllSignedAtoms();
		for (multimap<bool, Atom*, eqbool>::iterator q = all_tmp->begin(); q != all_tmp->end(); q++) {
		  if ((tmpel = getElt((*q).second->getId(), state, tr, time_step+1, 1))->lock == 1 
		      || tmpel->lock == 4) {
#ifdef DEBUG_DIAG 
		    cerr << (*q).second->getId() << " is post-locked !!\n";
#endif		    
		    lock_flag = true;
		    break;;
		  }
		}
		delete all_tmp;
		all_tmp = 0;
	      }

	      if (!lock_flag) {   /* if the current_transition is the right one. */
#ifdef DEBUG_DIAG
		cerr << "[immediateSuccessors] --> transition chosen : \n";
		curr_trans->showTransition();
#endif
		if (curr_trans->getType() == "Fail")
		  trans_rank = 1;
		else trans_rank = 0;
		if (current_rank + trans_rank <= rank /*) {*/
		    && ! checkForDeadEnd ((*at_trans)[m])) {  /* reduce diag with rank and dead end trans */	  
		    
#ifdef PROBAS
		  if (post) /* update states cover. */
		    tr->updateCover(state, curr_trans);
		  else tr->updateCover(pre_state, state, start_step, time_step, curr_trans);
#endif
		  //BEWARE => TODO : in case of equality... need a set of already used trans per state.
		  //if ((post & (curr_trans->proba <= 1 - state->cover)) | (pre & (curr_trans->proba <= 1 - pre_state->cover))) {
		  liketrans = new likelyTransition(curr_trans);
		  liketrans->val = curr_val->at;
		  liketrans->at = curr_at->at;
		  liketrans->start_step = start_step;
#ifdef PROBAS
		  /*liketrans->rank = pre_state->proba * curr_trans->proba;*/
		  liketrans->rank = (*state->trans_probs)[curr_trans->trans_id];
#else
		  liketrans->rank = current_rank + trans_rank;
#endif
		  likely_trans->insert(pair<double, likelyTransition*>(liketrans->rank, liketrans));	
		}
		else {
#ifdef PROBAS
		  if (post)  /* update states cover. */
		    tr->updateCover(state, curr_trans);
		  else tr->updateCover(pre_state, state, start_step, time_step, curr_trans);
#endif
#ifdef DEBUG_DIAG
		  cerr << "out of rank !\n";
		  cerr << "current rank : " << current_rank << " -- rank : " << rank << endl;
#endif
		}
		break_flag = true;     	/*set the break state var and the break flag. (only one statevar for each imSuc session). */
		if (curr_stv->getType() == "Command" && cmd_modevar == NULL) { 
		  cmd_modevar = curr_trans->modevar;
		}
	      }
	      else {
#ifdef DEBUG_DIAG
		cerr << "[immediateSuccessors] --> transition impossibility (" << time_step << ") :\n";
		curr_trans->showTransition();
#endif
	      }
	  }
	}
      }
    }
    if (break_flag) {
      break;
    }
#ifdef HANDLE_TCM
    /* commands do not need to be supported by transitions. 
       this 'hack' only works on commands single conflicts. */
    else if (curr_val->prevTrans != NULL) { 
      //else if (curr_val->prevTrans_time != -1) {
      if (curr_stv->type == "Command" && soln->size() == 1 && curr_val->prevTrans->post == NULL) {// & curr_val->prevTrans_time != -1) {
#ifdef DEBUG_DIAG
	  cerr << "rerouting sol at " << time_step << endl;
#endif
	curr_val->prevTrans = NULL;
	routed_traj = new Trajectory(tr, tr->getCurrentTime()+1);
	n_trajectories->insert(pair<double, Trajectory*>(tr->rank, routed_traj));
#ifdef DEBUG_DIAG
	  cerr << "rerouted trajectory :\n";
	  routed_traj->showTrajectoryTransitions();
#endif
	break;
      }
    }
#endif
  }
  delete stas;
  stas = 0;
  transitions_checked = transitions_checked + likely_trans->size();
  for (multimap<double, likelyTransition*, greater<double> >::iterator lt = likely_trans->begin(); lt != likely_trans->end(); lt++) {
    curr_trans = (*lt).second->trans;
    n_trajectories = constructSuccessors((*lt).second, tr, time_step, n_trajectories, rank);
    likely_trans->erase(lt);
  }
  
  for (multimap<double, Trajectory*, greater<double> >::iterator t = n_trajectories->begin(); t != n_trajectories->end(); t++) {
    /* get the next occurence of the current state var in the trajectory. Not sure it works with tCM... */
    if (time_step != (*t).second->getCurrentTime()) {
      if ((next_step = (*t).second->getNext(curr_stv, time_step)) == 0) {
	/* add the 'missing' component to the last state of the trajectory 
	 * (due to partial path reconstruction from observation). */
	(*t).second->addComp(curr_stv->getComp(), (*t).second->getState((*t).second->getCurrentTime()), (*t).second->getCurrentTime(), 1, 0);
#ifdef USE_BDD
	addToStateBdd((*t).second->getState((*t).second->getCurrentTime()), curr_stv->getComp()->getName());
#endif	
	next_step = (*t).second->getCurrentTime();
      }
    }
    else next_step = 0;
    
#ifdef DEBUG_DIAG
      (*t).second->showTrajectoryTransitions();
      cerr << endl;
#endif
    
    if (!(conf_post_prev = getConflicts((*t).second, time_step, true))->empty()) {
      trajs = immediateSuccessors(conf_post_prev, (*t).second, time_step, rank, trajs, true);
      for (hash_map<const char*, disConflict*, hash<const char*>, eqstr>::iterator itd = conf_post_prev->begin(); itd != conf_post_prev->end(); itd++) {
	delete (*itd).second;
	(*itd).second = 0;
      }
      delete conf_post_prev;
      conf_post_prev = 0;
    }
    else if (!(conf_pre_next = getConflicts((*t).second, time_step, false))->empty()) {
      trajs = immediateSuccessors(conf_pre_next, (*t).second, time_step, rank, trajs, false);
      for (hash_map<const char*, disConflict*, hash<const char*>, eqstr>::iterator itd = conf_pre_next->begin(); itd != conf_pre_next->end(); itd++) {
	delete (*itd).second;
	(*itd).second = 0;
      }
      delete conf_pre_next;
      conf_pre_next = 0;
    }
    else if (next_step != 0 && !(conf_post_next = getConflicts((*t).second, next_step, false))->empty()) {
      trajs = immediateSuccessors(conf_post_next, (*t).second, next_step, rank, trajs, false);
      for (hash_map<const char*, disConflict*, hash<const char*>, eqstr>::iterator itd = conf_post_next->begin(); itd != conf_post_next->end(); itd++) {
	delete (*itd).second;
	(*itd).second = 0;
      }
      delete conf_post_next;
      conf_post_next = 0;
    }
    else if (conf_post_prev->empty() 
	     && conf_pre_next->empty() 
	     && (next_step == 0 || conf_post_next->empty())) {
#ifdef DEBUG_DIAG
      cerr << "Traj added to immSuccs final results...\n";
#endif
#ifdef PROBAS
      (*t).second->updateStatesProb ();
      trajs->insert(pair<double, Trajectory*>((*t).second->getLastState ()->proba, (*t).second));
#else
      trajs->insert(pair<double, Trajectory*>((*t).second->rank, (*t).second));
#endif
      delete_flag = false; /* useless ? */
      delete conf_post_next;
      conf_post_next = 0;
      delete conf_pre_next;
      conf_pre_next = 0;
      delete conf_post_prev;
      conf_post_prev = 0;
    }
  }
  
  delete n_trajectories;
  n_trajectories = 0;
  delete likely_trans;
  likely_trans = 0;
  delete conf_pre_next;
  conf_pre_next = 0;
  delete conf_post_prev;
  conf_post_prev = 0;
  delete conf_post_next;
  conf_post_next = 0;
  delete tr;
  tr = 0;
  return trajs;
}

/**
 * Get transition conflicts. This method performs a full search, i.e. over the entire
 * state. It could be reduced.
 * 
 * @param tr is the trajectory over which conflicts has to be found.
 * @param time_step is the current time step the system is working on.
 * @param post defines if the search for transition conflicts is made on pre or post.
 * @result a set of single elts that are part of transition conflicts.
 */
hash_map<const char*, disConflict*, hash<const char*>, eqstr>* Diagnoser::getConflicts(Trajectory* tr, int time_step, bool post) {
  /*int curr_trans_time;*/
  State* current_state = tr->getState(time_step);
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* res = new hash_map<const char*, disConflict*, hash<const char*>, eqstr>();
  Transition* trans;
  vector<Atom*>* trans_post;
  TrajEl* curr_el;
  int conf_pos;
  vector<Transition*>* store_trans = new vector<Transition*>();
  
  if (post) {
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator curr_it = current_state->state->begin(); 
	 curr_it != current_state->state->end(); curr_it++) {
      if ((*curr_it).second->prevTrans != NULL && !inVector((*curr_it).second->prevTrans, store_trans)) {
	trans = (*curr_it).second->prevTrans;
	store_trans->push_back(trans);
	/* in case trans is not a multiple commands transition. */
	if (trans->post != NULL) {
	  trans_post = trans->post->getAllAtoms();
	  for (unsigned int i=0; i<trans_post->size(); i++) {
	    curr_el = (*current_state->state)[(*trans_post)[i]->getId()];
	    if (curr_el->label == "false") {
	      /* get the last time occurence of the elt. */
	      tr->getElt(curr_el->getId(), time_step - 1, 0, conf_pos);
	      (*res)[curr_el->getId()] = new disConflict(curr_el, conf_pos);
#ifdef DEBUG_DIAG
		cerr << "[getConflicts()] --> elts of the POST NORMAL conflict (" << time_step <<") :\n";
		cerr << curr_el->label + " : " << curr_el->getId() << endl;
		cerr << "conf_pos : " << conf_pos << endl;
		cerr << "[getConflicts()] --> post conflict detected on :\n";
		trans->showTransition();
#endif
	    }
	  }
	  delete trans_post;
	  trans_post = 0;
	}
#ifdef HANDLE_TCM
	/* in case trans is a multiple commands transition. BEWARE : Return the current conflict even if it is not really a conflict, i.e. its value could be valid... To be changed.*/
	else {
	  curr_trans_time = time_step - (*curr_it).second->prevTrans_time;
	  /* get the 'normal' transition conflicts. */
	  trans_post = (*trans->st_post)[curr_trans_time]->getAllAtoms();
	  for (unsigned int i=0; i<trans_post->size(); i++) {
	    curr_el = (*current_state->state)[(*trans_post)[i]->getId()];
	    if (curr_el->label == "false" && curr_el->prevTrans != NULL) {
	    /* get the last time occurence of the elt. */
	    tr->getElt(curr_el->getId(), time_step - 1, 0, conf_pos);
	    (*res)[(*trans_post)[i]->getId()] = new disConflict(curr_el, conf_pos);
#ifdef DEBUG_DIAG
		cerr << "[getConflicts()] --> elts of the POST TCM conflict (" << time_step <<") :\n";
		cerr << "false : " << (*trans_post)[i]->getId() << " -- " << curr_el->lock << endl;
		cerr << "conf_pos : " << conf_pos << endl;
		cerr << "[getConflicts()] --> post conflict detected on :\n";
		trans->showTransition();
#endif
	    }
	  }
	  delete trans_post;
	  trans_post = 0;
	}
#endif
      }
    }
  }
  /* PRE conflict detection. */
  else {
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator curr_it2 = current_state->state->begin(); curr_it2 != current_state->state->end(); curr_it2++) {
      if ((*curr_it2).second->prevTrans != NULL) {
	trans = (*curr_it2).second->prevTrans;
	if (trans->post != NULL) { //no TCM only.
	  for (multimap<bool, Atom*, eqbool>::iterator j = trans->pre->atoms->begin(); 
	       j != trans->pre->atoms->end(); j++) {
	    curr_el = tr->getElt((*j).second->getId(), time_step - 1, 0, conf_pos);
	    if (((*j).first && curr_el->label == "false") || (!(*j).first && curr_el->label == "true")) {
	      (*res)[(*j).second->getId()] = new disConflict(curr_el, conf_pos, (*j).first);
#ifdef DEBUG_DIAG
		cerr << "[getConflicts()] --> elt of the PRE NORMAL conflict (" << time_step <<") :\n";
		cerr << "elt of the prevTrans : " << (*curr_it2).second->getId() << endl;
		cerr << "conf_pos : " << conf_pos << endl;
		cerr << "false : " << (*j).second->getId() << endl;
		cerr << "[getConflicts()] --> pre conflict detected on :\n";
		trans->showTransition();
#endif
	    }
	  }
	}   
      }
    }
  }
#ifdef DEBUG_DIAG
    cerr << "******* Showing conflicts *******\n";
    cerr << "post (non-opt) : " << post << " - time step : " << time_step << endl;
    viewHashMapConflicts(res);
#endif
    delete store_trans;
    store_trans = 0;
    return res;
}

bool Diagnoser::checkForDeadEnd (Transition* trans) {
  Transition* tp;
  if (! trans->commanded && trans->type == "Fail") {
    if (isNull (tp = (*redux_diag)[(int) trans])) {
      (*redux_diag)[(int) trans] = trans;
      return false;
    }
    else return true;
  }
  else return false;
}

bool Diagnoser::checkForDeadEndExt (Transition* trans, Trajectory* tr) {
  Transition* tp;
  if (! trans->commanded && trans->type == "Fail") {
    if (isNull (tp = (*redux_diag)[(int) trans])) {
      if (! isNull (tr->trans_diag)) {
	for (multimap<int, Transition*>::iterator it = tr->trans_diag->begin ();  /* extended check */
	     it != tr->trans_diag->end (); it++) {
	  if (! isNull (tp = (*redux_diag)[(int) (*it).second]))
	    return true;
	}
      }
      return false;
    }
    else return true;
  }
  else {
    if (! isNull (tr->trans_diag)) {
      for (multimap<int, Transition*>::iterator it = tr->trans_diag->begin ();  /* extended check */
	   it != tr->trans_diag->end (); it++) {
	if (! isNull (tp = (*redux_diag)[(int) (*it).second]))
	  return true;
      }
    }
    return false;
  }
}

void Diagnoser::addResult (Trajectory* res_tr, multimap<double, Trajectory*, greater<double> >* res, double &norm) {
  /*Transition* tp;*/
  
  /*for (multimap<int, Transition*>::iterator it = res_tr->trans_diag->begin ();
       it != res_tr->trans_diag->end (); it++) {
    if (! (*it).second->commanded && (*it).second->type == "Fail") {
      if (! isNull (tp = (*redux_diag)[(int) (*it).second]))
	return;
      (*redux_diag)[(int) (*it).second] = (*it).second;
    }
  }*/

#ifdef PROBAS
  res_tr->updateStatesProb();
  norm += res_tr->getLastState()->proba;
  res->insert(pair<double, Trajectory*>(res_tr->getLastState()->proba, res_tr));
#else	      
  res->insert(pair<double, Trajectory*>(res_tr->rank, res_tr));
#endif
}

/**
 * COVER TRACK.
 * This procedure performs a search over a trajectory which is no more consistent with observations.
 * It incrementally constructs new consistent trajectories and return them as soon as they are found.
 * The search is performed in a semi width / full depth manner :
 * - width is limited by using rank (1 for failure, 0 for nominal).
 * - each newly constructed trajectory is tracked for consistence until it is found to be a solution or it is killed.
 * We try to start from any trajectory... (not only from the nominal one).
 *
 * @param trajec is the trajectory from which the search is started.
 * @param rank_lim_low is the low rank limit of the search.
 * @param rank_lim_high is the high rank limit of the search.
 * @param obs is the set of observable at current time step.
 * @param amp is the amplitude of the search. it is not used yet.
 * @result a set of trajectories consistent with observation through their entire history.
 */
multimap<double, Trajectory*, greater<double> >* Diagnoser::coverTrack2(Trajectory* trajec, double rank_lim_low, double rank_lim_high, multimap<bool, Atom*, eqbool>* obs, int amp, double &current_rank) {
  clock_t start, end = 0;
  double elapsed;
  bool first = false;
  start = clock();
  transitions_checked = 0;
  trajectories_checked = 0;
  double rank = rank_lim_low;
#ifdef PROBAS
  double norm = 0; /* normalize trajectories probabilities. */
#endif
  int time_step = trajec->getCurrentTime();
  multimap<double, Trajectory*, greater<double> >* temp_trajs; //store new trajectories before adding them to trajs.
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* conf;
  stack<Trajectory*>* trajs;
  multimap<double, Trajectory*, greater<double> >* trajs_init;
  multimap<double, Trajectory*, greater<double> >* res = new multimap<double, Trajectory*, greater<double> >();
  int i;
  int count_sons;
  bool speedy = false;
  Trajectory* curr_traj;
  Trajectory* trajec_copy;
  while (res->empty() && rank <= rank_lim_high) {
    end = clock();
    cerr << "Rank : " << rank << "(" << ((double)(end - start))/CLOCKS_PER_SEC << ")\n";
    trajs = new stack<Trajectory*>();
    trajec_copy = new Trajectory(trajec, trajec->getCurrentTime () + 1);
    trajs_init = observationResults(obs, trajec_copy, time_step, rank, 0);
    for (multimap<double, Trajectory*, greater<double> >::iterator o = trajs_init->begin(); o != trajs_init->end(); o++) {
      //trajs->push(new Trajectory((*o).second, time_step+1));
      trajs->push((*o).second);
    }
    while (!trajs->empty()) {
      speedy = false;
      count_sons = 1;
      first = true;
      curr_traj = trajs->top();
      trajs->pop();
      i = curr_traj->consist_time;
      while (count_sons>0 && i>0) {
      //cerr << "amp : " << amp << endl;
      //while (count_sons>0 & i>(current_time-amp-1)) {
	temp_trajs = new multimap<double, Trajectory*, greater<double> >();
	conf = getConflicts(curr_traj, i, true);
	if (!conf->empty()) {
	  temp_trajs = immediateSuccessors(conf, curr_traj, i, rank, temp_trajs, true);
#ifdef DEBUG_DIAG
	  cerr << "result size from immediate succs : " << temp_trajs->size() << endl;
	  compareTrajs(temp_trajs);
#endif
	  for (multimap<double, Trajectory*, greater<double> >::iterator l = temp_trajs->begin(); l != temp_trajs->end(); l++) {
	    if ( i != 1) {
	      (*l).second->consist_time--;
	      trajs->push((*l).second);
	      count_sons++;
	    }
	    else {
	      end = clock();
	      //(*l).second->showTrajectoryLight();
	      (*l).second->showTrajectoryTransitions();
	      cerr << "Time : " << ((double)(end - start))/CLOCKS_PER_SEC << endl;
	      /* update states probabilities. */
#ifdef PROBAS
	      (*l).second->updateStatesProb();
	      norm += (*l).second->getLastState()->proba;
	      res->insert(pair<double, Trajectory*>((*l).second->getLastState()->proba, (*l).second));
#else
	      res->insert(pair<double, Trajectory*>((*l).second->rank, (*l).second));
#endif
	      unlockTrajectory((*l).second);
	    }
	  }
	  /*delete curr_traj;
	  curr_traj = 0;*/
	  count_sons--;
	  if (!trajs->empty() & count_sons>0) { 
#ifdef DEBUG_DIAG
	      cerr << "poping traj :\n";
	      cerr << "consist time : " << trajs->top()->consist_time << endl;
	      trajs->top()->showTrajectoryTransitions();
#endif
	    curr_traj = trajs->top();
	    trajs->pop();
	  }
	  else {
	    count_sons = 0;
	  }
	}
	else if (i != 1) {
	  curr_traj->consist_time--;
	}
	else {
	  end = clock();
	  //curr_traj->showTrajectoryLight();
	  curr_traj->showTrajectoryTransitions();
	  cerr << "Time : " << ((double)(end - start))/CLOCKS_PER_SEC << endl;
#ifdef PROBAS
	  /* update probabilities. */
	  curr_traj->updateStatesProb();
	  norm += curr_traj->getLastState()->proba;
	  res->insert(pair<double, Trajectory*>(curr_traj->getLastState()->proba, curr_traj));
#else
	  res->insert(pair<double, Trajectory*>(curr_traj->rank, curr_traj));
#endif
	  unlockTrajectory(curr_traj);
	  count_sons--;
	  //cerr << "count_sons : " << count_sons << endl;
	  if (!trajs->empty() & count_sons > 0) {
	    curr_traj = trajs->top();
	    trajs->pop();
	  }
	}
	delete conf;
	conf = 0;
	delete temp_trajs;
	temp_trajs = 0;
	
	if (count_sons != 0)
	  i = curr_traj->consist_time;
      }//end of time loop.	
    }//end of trajs loop.
    delete trajs;
    trajs = 0;

    rank++;
    //rans += 0.05;
    /*for (multimap<double, Trajectory*, greater<double> >::iterator rr = trajs_init->begin(); rr != trajs_init->end(); rr++)
      delete (*rr).second;
    delete trajs_init;*/
  }
  /* normalize trajectories probabilities. */
#ifdef PROBAS
  if (res != 0) {
    for (multimap<double, Trajectory*, greater<double> >::iterator itp = res->begin(); itp != res->end(); itp++)
      (*itp).second->proba = (*itp).second->getLastState()->proba / norm;
  }
#endif
  end = clock();
  delete trajec;
  trajec = 0;
  current_rank = rank - 1;
  elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
  /* view solutions. */
  cerr << "res's size : " << res->size() << endl;
  viewTrajectoriesTransitionsMultimap(res);
  cerr << "transitions checked : " << transitions_checked << endl;
  cerr << "trajectories created : " << Trajectory::traj_created << endl;
  cerr << "trajectories destroyed : " << Trajectory::traj_del << endl;
  cerr << "states created : " << State::state_created << endl;
  cerr << "states deleted : " << State::state_del << endl;
  cerr << "elts created : " << TrajEl::trajel_created << endl;
  cerr << "elts deleted : " << TrajEl::trajel_del << endl;
  updateTrajectoriesCover();
  cerr << "trajectories cover : " << traj_cover << endl;
  /*for (hash_map<int, double, hash<int> >::iterator cov_it = states_cover->begin(); cov_it != states_cover->end(); cov_it++)
    cerr << (*cov_it).first << " -- " << (*cov_it).second << endl;*/
  cerr << "cover_track execution time : " << elapsed << endl;
  //compareTrajs(res);
  return res;
}

/**
 * COVER TRACK3
 */
multimap<double, Trajectory*, greater<double> >* Diagnoser::coverTrack3(Trajectory* trajec, double rank_lim_low, double rank_lim_high, multimap<bool, Atom*, eqbool>* obs, int amp, double &current_rank) {
  clock_t start, end;
  double elapsed;
  start = clock();
  transitions_checked = 0;
  trajectories_checked = 0;
  double rank = rank_lim_low;
/*#ifdef PROBAS*/
  double norm = 0; /* normalize trajectories probabilities. */
/*#endif*/
  int time_step = trajec->getCurrentTime();
  multimap<double, Trajectory*, greater<double> >* temp_trajs; //store new trajectories before adding them to trajs.
  hash_map<const char*, disConflict*, hash<const char*>, eqstr>* conf;
  multimap<double, Trajectory*, greater<double> >* trajs;
  multimap<double, Trajectory*, greater<double> >* trajs_init;
  multimap<double, Trajectory*, greater<double> >* res = new multimap<double, Trajectory*, greater<double> >();
  int i;
  Trajectory* curr_traj = 0;
  Trajectory* trajec_copy;
  while (res->empty() && rank <= rank_lim_high) {
    end = clock();
    cerr << "Rank : " << rank << "/" << rank_lim_high << " (" << ((double)(end - start))/CLOCKS_PER_SEC << ")\n";
    trajs = new multimap<double, Trajectory*, greater<double> >();
    trajec_copy = new Trajectory(trajec, trajec->getCurrentTime () +1);
    trajs_init = observationResults(obs, trajec_copy, time_step, rank, 0);
    for (multimap<double, Trajectory*, greater<double> >::iterator o = trajs_init->begin(); o != trajs_init->end(); o++) {
#ifdef PROBAS
      trajs->insert (pair<double, Trajectory*>((*o).second->getLastState ()->proba, (*o).second));
#else
      trajs->insert (pair<double, Trajectory*>((*o).second->rank, (*o).second));
#endif    
    }
    while (!trajs->empty() || curr_traj != 0) {
      if (curr_traj == 0) {
	curr_traj = (*trajs->begin()).second;
	trajs->erase(trajs->begin());
      }
      i = curr_traj->consist_time;
      temp_trajs = new multimap<double, Trajectory*, greater<double> >();
      conf = getConflicts(curr_traj, i, true);
      if (!conf->empty()) {
	temp_trajs = immediateSuccessors(conf, curr_traj, i, rank, temp_trajs, true);
	for (multimap<double, Trajectory*, greater<double> >::iterator l = temp_trajs->begin(); l != temp_trajs->end(); l++) {
	  if (i != (*l).second->low_time+1) {
	    (*l).second->consist_time--;
#ifdef PROBAS
	    trajs->insert(pair<double, Trajectory*>((*l).second->getLastState ()->proba, (*l).second));
#else
	    trajs->insert(pair<double, Trajectory*>((*l).second->rank, (*l).second));
#endif	
	  }
	  else {
	    end = clock();
	    cerr << "Time : " << ((double)(end - start))/CLOCKS_PER_SEC << endl;
	    if ((*l).second->rank == rank) {
/*#ifdef PROBAS
	      (*l).second->updateStatesProb();
	      norm += (*l).second->getLastState()->proba;
	      res->insert(pair<double, Trajectory*>((*l).second->getLastState()->proba, (*l).second));
#else	      
	      res->insert(pair<double, Trajectory*>((*l).second->rank, (*l).second));
#endif*/
	      addResult ((*l).second, res, norm);
	      /*(*l).second->showTrajectoryTransitions ();*/
	      /*unlockTrajectory((*l).second);*/
	    }
	    else { /* search finishes here ! */
	      cerr << "too high rank found: " << (*l).second->rank << endl;
	      curr_traj = 0;
	      for (multimap<double, Trajectory*, greater<double> >::iterator rr = temp_trajs->begin(); rr != temp_trajs->end(); rr++) {
		delete (*rr).second;
		(*rr).second = 0;
	      }
	      delete temp_trajs;
	      temp_trajs = 0;
	      delete conf;
	      conf = 0;
	      for (multimap<double, Trajectory*, greater<double> >::iterator it = trajs->begin(); it != trajs->end(); it++) {
		delete (*it).second;
		(*it).second = 0;
	      }
	      delete trajs;
	      trajs = 0;
	      delete trajs_init;
	      trajs_init = 0;
	      goto end;
	    }
	  }
	}
	curr_traj = 0;
      }
      else if (i != curr_traj->low_time+1) {
	curr_traj->consist_time--;
      }
      else {
	end = clock();
	cerr << "Time : " << ((double)(end - start))/CLOCKS_PER_SEC << endl;
	if (curr_traj->rank == rank) {
/*#ifdef PROBAS
	  curr_traj->updateStatesProb();
	  norm += curr_traj->getLastState()->proba;
	  res->insert(pair<double, Trajectory*>(curr_traj->getLastState()->proba, curr_traj));
#else	  
	  res->insert(pair<double, Trajectory*>(curr_traj->rank, curr_traj));
#endif*/
	  addResult (curr_traj, res, norm);
	  /*curr_traj->showTrajectoryTransitions ();*/
	  /*unlockTrajectory(curr_traj);*/
	}
	else {
	  cerr << "too high rank found: " << curr_traj->rank << endl;
	  curr_traj = 0;
	  delete conf;
	  conf = 0;
	  for (multimap<double, Trajectory*, greater<double> >::iterator it = trajs->begin(); it != trajs->end(); it++) {
	    delete (*it).second;
	    (*it).second = 0;
	  }
	  delete trajs;
	  delete trajs_init;
	  trajs_init = 0;
	  goto end;
	}
	curr_traj = 0;
      }
      delete conf;
      conf = 0;
      delete temp_trajs;
      temp_trajs = 0;
    }/* end of trajs loop */
    delete trajs;
    trajs = 0;
    rank++;
  }
 end:
  /* normalize trajectories probabilities. */
#ifdef PROBAS
  if (res != 0) {
    for (multimap<double, Trajectory*, greater<double> >::iterator itp = res->begin(); itp != res->end(); itp++)
      (*itp).second->proba = (*itp).second->getLastState()->proba / norm;
  }
#endif
  end = clock();
  current_rank = rank - 1;
  elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
  cerr << "res's size : " << res->size() << endl;
  /*viewTrajectoriesMultimap(res);*/
  viewTrajectoriesTransitionsMultimap(res);
  /*viewTrajectoriesMultimapFull(res);*/
#ifdef DEBUG
  cerr << "transitions checked : " << transitions_checked << endl;
  cerr << "trajectories created : " << Trajectory::traj_created << endl;
  cerr << "trajectories destroyed : " << Trajectory::traj_del << endl;
  cerr << "states created : " << State::state_created << endl;
  cerr << "states deleted : " << State::state_del << endl;
  cerr << "elts created : " << TrajEl::trajel_created << endl;
  cerr << "elts deleted : " << TrajEl::trajel_del << endl;
#endif
  updateTrajectoriesCover();
  /*cerr << "trajectories cover : " << traj_cover << endl;
    for (hash_map<int, double, hash<int> >::iterator cov_it = states_cover->begin(); cov_it != states_cover->end(); cov_it++)
    cerr << (*cov_it).first << " -- " << (*cov_it).second << endl;*/
  cerr << "cover_track execution time : " << elapsed << endl;
  redux_diag->clear ();
  delete trajec;
  return res;
}
/************************
 * Conflicts Management *
 ************************/
//add a conflict to the database.
/*void Diagnoser::addConflictToDB(bool post, const char* id, Transition* trans) {
  if ((*conflictsDB)[trans->trans_id] == NULL)
    (*conflictsDB)[trans->trans_id] = new pair<vector<const char*>*, vector<const char*>* >(new vector<const char*>(), new vector<const char*>());
    if (post & !inVector(id, (*conflictsDB)[trans->trans_id]->second)) {
    //cerr << "adding to conflictsDB : " << id << endl;;
    //trans->showTransition();
    (*conflictsDB)[trans->trans_id]->second->push_back(id);
  }
  else if (!inVector(id, (*conflictsDB)[trans->trans_id]->first)) {
    //cerr << "adding to conflictsDB : " << id << endl;
    //trans->showTransition();
    (*conflictsDB)[trans->trans_id]->first->push_back(id);
  }
}

//test for a conflict in the conflicts' database.
bool Diagnoser::inConflictsDatabase(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* pre_state, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, Transition* trans) {
  vector<const char*>* pre;
  vector<const char*>* post;
  bool flag = false;
  if (debug == 0) {
    debug = 1;
    flag = true;
  }
  //cerr << "trans_id : " << trans->trans_id << endl;
  if ((*conflictsDB)[trans->trans_id] != NULL) {
    pre = (*conflictsDB)[trans->trans_id]->first;
    post = (*conflictsDB)[trans->trans_id]->second;
    for (int i=0; i<pre->size(); i++) {
      if ((*pre_state)[(*pre)[i]]->label == "true") {
	printDebug(1, "conflict already in conflictsDB !!!");
	if (debug) {
	  cerr << (*pre)[i] << endl;
	  trans->showTransition();
	}
	if (flag) debug = 0;
	return true;
      }
    }
    for (int j=0; j<post->size(); j++) {
      if ((*state)[(*post)[j]]->label == "true") {
	printDebug(1, "conflict already in conflictsDB !!!");
	if (debug) {
	  cerr << (*post)[j] << endl;
	  trans->showTransition();
	}
	if (flag) debug = 0;
	return true;
      }
    if (flag) debug = 0;
    return false;
    }
  }
  else {
    //cerr << "transition is not in DB\n";
    if (flag) debug = 0;
    return false;
  }
  }*/

/*************************
 * Observation Management*
 *************************/

/**
 * Insert observations into a trajectory.
 * 
 * @param obs is the set of observations.
 * @param tr is the trajectory to be updated.
 * @param rank is the current rank of the search.
 * @result is a set of trajectories updated with incoming observations.
 */
/* DEPRECATED: shouldn't be used anymore */
multimap<double, Trajectory*, greater<double> >* Diagnoser::observationResults(vector<Atom*>* obs, Trajectory* tr, int time_step, double rank, bool nom) {
  cerr << "inserting observations at time_step (deprecated version): " << time_step << endl;
  lock_comp = 1;
  multimap<double, Trajectory*, greater<double> >* res = new multimap<double, Trajectory*, greater<double> >();
  Trajectory* traj_res;
  State* curr_st;
  vector<ItmsRes*>* itms_tmp;
  State* current_state = new State(tr->getState(time_step), 1);
  /* set observable flag. */
  current_state->observable = true;
#ifdef RECONF
  current_state->modifs->clear();
#endif
  multimap<bool, TrajEl*, eqbool>* add_props = new multimap<bool, TrajEl*, eqbool>();
  for (unsigned int i=0; i<obs->size(); i++) {
    //lockTrajEls(getElt((*obs)[i]->getId(), current_state, tr, time_step), current_state);
    lockTrajElsPerm(getElt((*obs)[i]->getId(), current_state, tr, time_step), current_state);
    add_props->insert(pair<bool, TrajEl*>(1, (*current_state->state)[(*obs)[i]->getId()]));
  }
  viewHashMapTrajElLow (current_state->state);
  itms_tmp = addPropositions(add_props, NULL, current_state, NULL, nom, rank, tr->rank, tr, time_step, 1);
  cerr << "itms_tmp size: " << itms_tmp->size () << endl;
  for (unsigned int u1=0; u1<itms_tmp->size(); u1++) {
    curr_st = (*itms_tmp)[u1]->res;
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator u2 = (*itms_tmp)[u1]->already_changed->begin(); u2 != (*itms_tmp)[u1]->already_changed->end(); u2++) {
      if ((*u2).second->getLabel() != "false")
	lockTrajEls((*u2).second, curr_st);
#ifdef RECONF
      else
	curr_st->modifs->push_back((*u2).second->at);
#endif
    }
    traj_res = new Trajectory(tr, time_step);
    current_state = (*itms_tmp)[u1]->res;
    traj_res->addState(time_step, current_state);
#ifdef PROBAS
    /* compute probability of the state. */
    traj_res->computeStateProb(time_step); //useless
    res->insert(pair<double, Trajectory*>(traj_res->getState(time_step)->proba, traj_res));
#else
    res->insert(pair<double, Trajectory*>(traj_res->getState(time_step)->rank, traj_res));
#endif
  }
  
  for (unsigned int dd = 0; dd < itms_tmp->size(); dd++) {
    delete (*itms_tmp)[dd];
    (*itms_tmp)[dd] = 0;
  }
  delete itms_tmp;
  itms_tmp = 0;
  delete tr;
  tr = 0;
  delete add_props;
  add_props = 0;
  cerr << "results from observations computed : " << res->size() << endl;
  for (multimap<double, Trajectory*, greater<double> >::iterator z = res->begin(); z != res->end(); z++) {
#ifdef PROBAS
    cerr << "posterior proba on obs : " << (*z).second->getState(time_step)->obs_cond_prob << endl;
    cerr << "proba : " << (*z).second->getState(time_step)->proba << endl;
    cerr << "cover : " << (*z).second->getState(time_step)->cover << endl;
#endif
    viewHashMapTrajElLow((*z).second->getState(time_step)->state);
    cerr << endl;
  }
  lock_comp = 0;
  return res;
}

multimap<double, Trajectory*, greater<double> >* Diagnoser::observationResults(multimap<bool, Atom*, eqbool>* obs, Trajectory* tr, int time_step, double rank, bool nom) {
  cerr << "inserting observations at time_step : " << time_step << endl;
  /*lock_comp = 1;*/  /* ??? */
  double obs_norm = 0;
  multimap<double, Trajectory*, greater<double> >* res = new multimap<double, Trajectory*, greater<double> >();
  Trajectory* traj_res;
  State* curr_st;
  vector<ItmsRes*>* itms_tmp;
  State* current_state = new State(tr->getState(time_step), 1);
  current_state->observable = true; /* set observable flag. */
#ifdef RECONF
  current_state->modifs->clear();
#endif
  multimap<bool, TrajEl*, eqbool>* add_props = new multimap<bool, TrajEl*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = obs->begin (); it != obs->end (); it++) {
    lockTrajElsPerm(getElt((*it).second->getId(), current_state, tr, time_step), current_state, !(*it).first);
    /*lockTrajEls(getElt((*it).second->getId(), current_state, tr, time_step), current_state);*/
    add_props->insert(pair<bool, TrajEl*>((*it).first, (*current_state->state)[(*it).second->getId()]));
  }
  itms_tmp = addPropositions(add_props, NULL, current_state, NULL, nom, rank, tr->rank, tr, time_step, 0);

#ifdef PROBAS
  /* probas normative cycle */
  for (unsigned pu=0; pu<itms_tmp->size (); pu++)
    obs_norm += (*itms_tmp)[pu]->res->obs_cond_prob;
#endif

  for (unsigned int u1=0; u1<itms_tmp->size(); u1++) {
    curr_st = (*itms_tmp)[u1]->res;
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator u2 = (*itms_tmp)[u1]->already_changed->begin(); 
	 u2 != (*itms_tmp)[u1]->already_changed->end(); u2++) {
      if ((*u2).second->getLabel() != "false")
	lockTrajEls((*u2).second, curr_st);
	/*lockTrajElsPerm((*u2).second, curr_st);*/
#ifdef RECONF
      else
	curr_st->modifs->push_back((*u2).second->at);
#endif
    }
    traj_res = new Trajectory(tr, time_step);
    /*current_state = (*itms_tmp)[u1]->res;*/
    traj_res->addState(time_step, curr_st);
#ifdef PROBAS
    /* compute probability of the state. */
    traj_res->computeStateProb(time_step); //useless
    curr_st->obs_cond_prob = curr_st->obs_cond_prob / obs_norm;
    traj_res->computeStateProb (time_step);
    res->insert(pair<double, Trajectory*>(traj_res->getState(time_step)->obs_cond_prob, traj_res));
#else
    res->insert(pair<double, Trajectory*>(traj_res->getState(time_step)->rank, traj_res));
#endif
  }
  
  for (unsigned int dd = 0; dd < itms_tmp->size(); dd++) {
    delete (*itms_tmp)[dd];
    (*itms_tmp)[dd] = 0;
  }
  delete itms_tmp;
  itms_tmp = 0;
  /*delete tr;
  tr = 0;*/  /* BUGGY... ??! */
  delete add_props;
  add_props = 0;
  cerr << "results from observations computed : " << res->size() << endl;
  for (multimap<double, Trajectory*, greater<double> >::iterator z = res->begin(); z != res->end(); z++) {
#ifdef PROBAS
    /*cerr << "posterior proba on obs : " << (*z).second->getState(time_step)->obs_cond_prob << endl;
    cerr << "proba : " << (*z).second->getState(time_step)->proba << endl;
    cerr << "cover : " << (*z).second->getState(time_step)->cover << endl;*/
#endif
    /*viewHashMapTrajElLow((*z).second->getState(time_step)->state);*/
    cerr << endl;
  }
  lock_comp = 0;
  return res;
}

/* view an hash_map of conflicts. */
void Diagnoser::viewHashMapConflicts(hash_map<const char*, disConflict*, hash<const char*>, eqstr>* confs) {
  for (hash_map<const char*, disConflict*, hash<const char*>, eqstr>::iterator it = confs->begin(); it != confs->end(); it++) {
    cerr << (*it).first << " -- " << (*it).second->conf->label << " -- " << (*it).second->conf->lock;
#ifdef PROBAS      
    cerr << " -- " << (*it).second->conf->proba << endl;
#else
    cerr << endl;
#endif
  }
}
