/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Scheduler.h"

//#define VERB_SCHED

using namespace std;

bool running_status = 0;
double Scheduler::simTime = 0;
#ifdef DISCRETE_ONLY
bool discretemdl = 0;
#endif

/** 
 * constructor with flag for debug.
 */
Scheduler::Scheduler(string mn) : Diagnoser () { 
  model_name = mn;
  running_status = 1;
  current_rank = 0;
  contEvents = new multimap<bool, Atom*, eqbool>();
  comEvents = new vector<Atom*>();
  if (initSchD() != 1)
    exit(1);
}
/* destructor. */
Scheduler::~Scheduler() {
  delete Eagenda;
  delete contEvents;
  delete comEvents;
}

/**
 * Initialize the scheduler, i.e. launch the agenda of event, an maybe more later ;-).
 */
int Scheduler::initSchD() {
  discrete_time = 0;
  Eagenda = new map<double, scenarEvent*>();
#ifdef BENCHMK
  /* open/create benchmark file. */
  /*char buf[15];
  char* s;
  s = buf;
  strcpy(s, model_name.c_str());
  strcat(s, ".bmk1");
  ofstream benchmk_stream(s, ios::trunc);
  cerr << "[Scheduler] --> benchmark file ok.\n";*/
#endif
  return 1;
}
//set the running status of the scheduler
void Scheduler::setRunningStatus(bool st) {
  running_status = st;
}
/** 
 * add a set of commands to the existing queue.
 * Also check for channels problem : if commands are not on the same channel...
 * atoms that are not commands can be added too...
 * @param com is the incoming command.
 * @param start_time is the time the command is required to be started.
 */
void Scheduler::addCommandEvent(Atom* com, int start_time) {
#ifdef VERB_SCHED
  cerr << "adding command : " << com->getId() << " -- " << start_time << endl;
#endif
  if ((*Eagenda)[(double) start_time] == NULL)
    (*Eagenda)[(double)start_time] = new scenarEvent(true, new vector<Atom*>());
  else {
    vector<Atom*>* curr_set = (*Eagenda)[(double)start_time]->incomingEvent;
    while (!intersectClauses(com, curr_set)) {
      start_time++;
      if ((*Eagenda)[(double)start_time] == NULL) {
	(*Eagenda)[(double)start_time] = new scenarEvent(true, new vector<Atom*>());
	break;
      }
      else
	curr_set = (*Eagenda)[(double)start_time]->incomingEvent;
    }
  }
  (*Eagenda)[(double)start_time]->command = true;
  (*Eagenda)[(double)start_time]->incomingEvent->push_back(com);
}
/**
 * organize the commands into the agenda, i.e. defragment ;)
 * @param time_step is the current time step.
 * @param increase specifies if it is authorized to add commands to the current time.
 * @returnreturn internal_com boolean flag to true if next command is internal, false otherwise. 
 */
bool Scheduler::organizeCommands(int time_step, bool increase) {
  scenarEvent* inco_scenar;
  bool internal_com = false;
  int pos;
  /* search for the next stored command. */
  for (map<double, scenarEvent*>::iterator it = Eagenda->upper_bound((double) time_step); it != Eagenda->end(); it++) {
    if ((*it).second != NULL) {
      pos = (int) (*it).first;
      /* move the elt to time_step + 1 if needed. */
      if (pos > time_step + 1) {
	(*Eagenda)[(double) (time_step + 1)] = (*it).second;
	(*Eagenda)[(double) (*it).first] = NULL;
      }
      /* leave. */
      internal_com = true;
      break;
    }
  }
  if (!internal_com) {
    inco_scenar = readFromScenario();
    if (!inco_scenar->incomingEvent->empty() && increase)
      (*Eagenda)[(double) (discrete_time + 1)] = inco_scenar;
    else if (!inco_scenar->incomingEvent->empty() && !increase)
      (*Eagenda)[(double) discrete_time] = inco_scenar;
  }
  return internal_com;
}
/**
 * Retrieve the none command, i.e. idle command into an element's model. 
 * It models waves-like flow of commands.
 * This is also used to constrain the diagnosis engine when searching to solution.
 * @param coms is a set of commands from which none values have to be retrieved.
 * @return the none command corresponding to the commands in coms.
 *
 */
vector<Atom*>* Scheduler::getNoneCommands(vector<Atom*>* coms) {
  vector<Atom*>* links;
  vector<Atom*>* res = new vector<Atom*>();
  for (unsigned int i=0; i<coms->size(); i++) {
    links = (*coms)[i]->getStateVar()->getAtomLinks();
    for (unsigned int j=0; j<links->size(); j++) {
      if ((*links)[j]->value == "none") {
	res->push_back((*links)[j]);
      }
    }
  }
  return res;
}

/**
 * Read incoming commands and observations from the scenario file.
 * @return the scenarEvent made of atoms in the scenario.
 */
scenarEvent* Scheduler::readFromScenario() {
  vector<Atom*>* res;
  scenarEvent* inco;
  Atom* at_tmp;
  bool command = false; //if the incoming event is a command or not.
  const char delimiters[] = " \n";
  char* lineptr;
  char* cp;
  char* token;
  lineptr = (char*)malloc(1024*sizeof(char));
  if (fgets(lineptr, 1024, scenar_stream) != NULL) {
    res = new vector<Atom*>();
    inco = new scenarEvent();
    cp = strdup(lineptr);
    token = strtok(cp, delimiters); //token for the type of the event.
    if (strcmp(token, "Command") == 0)
      command = true;
    while ((token = strtok(NULL, delimiters)) != NULL) {
      /*cerr << "regular token : " << token << endl;*/
      at_tmp = getAtom(token);
      if (at_tmp == NULL) {
	cerr << "Scheduler::readFromScenario --> received command is not valid : " + string(token) << endl;
	/*inco->command = command;
	inco->incomingEvent = res;*/
	return NULL;
      }
      else if (command) {
	cerr << "Scheduler receives a command: " + string(at_tmp->getId()) << endl;
	res->push_back(at_tmp);
      }
      else { //incoming events are observations.
	cerr << "Scheduler receives an observation: " + string(at_tmp->getId()) << endl;
	res->push_back(at_tmp);
      }
    }
  }
  else {
    cerr << "end of the scenario !\n";
    return NULL;
  }
  inco->command = command;
  inco->incomingEvent = res;
  return inco;
}

void Scheduler::loadScenario () {

  if ((scenar_stream = fopen(scenarFile, "r")) == NULL) {   /* open the scenario file. */
    cerr << "can't open scenario file: " << scenarFile << endl;
    return;
  }
  
  Atom* at_tmp;
  vector<Atom*>* res;
  bool command = false;
  double ev_date;
  const char delimiters[] = " \n";
  char *token, *cp;
  char* lineptr = (char*) malloc (1024*sizeof (char));
  
  while (fgets (lineptr, 1024, scenar_stream) != NULL) {
    cp = strdup(lineptr);
    res = new vector<Atom*>();
    token = strtok (cp, delimiters);  /* first token is the date */
    ev_date = atof (token);  /* TODO: migrate toward to strtol when time allows... */
    cerr << "event date: " << ev_date << endl;
    
    token = strtok(NULL, delimiters); /* token for the type of the event */
    if (strcmp(token, "Command") == 0)
      command = true;
    while ((token = strtok(NULL, delimiters)) != NULL) {
      cerr << "regular token : " << token << endl;
      at_tmp = getAtom(token);
      if (at_tmp == NULL) {
	cerr << "Scheduler::readFromScenario --> received command is not valid : " + string(token) << endl;
	return;
      }
      else if (command) {
	cerr << "Scheduler receives a command: " + string(at_tmp->getId()) << endl;
	res->push_back(at_tmp);
      }
      else { //incoming events are observations.
	cerr << "Scheduler receives an observation: " + string(at_tmp->getId()) << endl;
	res->push_back(at_tmp);
      }
    }
    (*Eagenda)[ev_date] = new scenarEvent (command, res);
  } /* end while fgets */
  fclose (scenar_stream);
}

/* set the post state and record next commands. Beware : no recursive formulas in post trans.*/
void Scheduler::triggerTransition(Transition* curr_trans, State* &current_step, State* &next_step, Trajectory* tra, hash_map<int, transitionCM*, hash<int> >* transitions_CM, int time_step) {
  StateFormula* post;
  multimap<bool, TrajEl*, eqbool>* add_props;
  TrajEl* curr_el;
  vector<ItmsRes*>* itms_res2;
  
  if (! discretemdl) 
    addBridgeTrans (curr_trans, current_rt);
  
  if (curr_trans->post != NULL) {                               /* first case is normal transition.*/
    post = curr_trans->post;
    add_props = new multimap<bool, TrajEl*, eqbool>();
    for (multimap<bool, Atom*, eqbool>::iterator l = post->atoms->begin(); 
	 l != post->atoms->end(); l++) {
      curr_el = getElt((*l).second->getId(), next_step, tra, time_step+1, 0);
      if (curr_el->getVarType() != "Command") {
	add_props->insert(pair<bool, TrajEl*>(1, curr_el));
	curr_el->prevTrans_time = time_step;                    /* update prevTrans. */
	curr_el->prevTrans = curr_trans;
      }
      else {                                                    /* add the command atoms to next commands set. */
	addCommandEvent((*l).second, time_step+1);
      }
    }
   
    itms_res2 = addPropositions(add_props, NULL, next_step, NULL, 1, tra->rank, tra->rank, tra, time_step+1, 0);
    next_step = (*itms_res2)[0]->res;
    
    if (! discretemdl) /* link changes to continuous level. */
      addBridgeConditions ((*itms_res2)[0]->already_changed);
    
    /* update prevTrans in itms_res2. */
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator al_it2 = (*itms_res2)[0]->already_changed->begin(); al_it2 != (*itms_res2)[0]->already_changed->end(); al_it2++) {
      if ((*al_it2).second->getLabel() == "false")
	(*al_it2).second->prevTrans = NULL;
    }
    delete (*itms_res2)[0];
    delete itms_res2;
    itms_res2 = 0;
    delete add_props;
    
    /* update nextTrans. */
    for (multimap<bool, Atom*, eqbool>::iterator m = curr_trans->pre->atoms->begin(); 
	 m != curr_trans->pre->atoms->end(); m++) {
      if ((*current_step->state)[(*m).second->getId()] == NULL) {  /* update elts in current step as needed. */
	tra->addComp((*m).second->getVarComp(), current_step, time_step, 1, 0);
#ifdef USE_BDD
	addToStateBdd(current_step, (*m).second->getVarComp()->getName());
#endif
      }
#ifdef VERB_SCHED
      cerr << (*m).second->getVarComp()->name << " : ";
#endif
      (*current_step->state)[(*m).second->getId()]->addToNextTrans(curr_trans);
    }
    curr_trans->showMove ();
  }
  else {   /* second case is transition with multiple commands. */
    /* get the command to track. */
    for (hash_map<int, StateFormula*, hash<int> >::iterator st_it = curr_trans->st_post->begin(); st_it != curr_trans->st_post->end(); st_it++) {
      post = (*st_it).second;
      add_props = new multimap<bool, TrajEl*, eqbool>();
      for (multimap<bool, Atom*, eqbool>::iterator l = post->atoms->begin(); l != post->atoms->end(); l++) {
	curr_el = getElt((*l).second->getId(), next_step, tra, time_step+1, 0);
	if ((*transitions_CM)[curr_trans->trans_id] == NULL) {  	/* update prevTrans if at next time step. */
	  curr_el->prevTrans = curr_trans;
	  curr_el->prevTrans_time = time_step;
	}
	if (curr_el->getVarType() != "Command") {
	  if (curr_el->prevTrans_time + (*st_it).first == time_step + 1) {
	    add_props->insert(pair<bool, TrajEl*>(1, curr_el));
	  }
	  /* patch for non commands variables at end of transitions. */
	  else
	    addCommandEvent((*l).second, time_step + (*st_it).first);}
	else {
	  addCommandEvent((*l).second, time_step + (*st_it).first);
	}
      }
      /* add a new transition to be tracked. */
      for (multimap<bool, Atom*, eqbool>::iterator cm = curr_trans->pre->atoms->begin(); cm != curr_trans->pre->atoms->end(); cm++)
	if ((*cm).second->getVarType() == "Command")
	  (*transitions_CM)[curr_trans->trans_id] = new transitionCM(curr_trans, time_step, (*cm).second); 
      
      itms_res2 = addPropositions(add_props, NULL, next_step, NULL, 1, tra->rank, tra->rank, tra, time_step+1, 0);
      next_step = (*itms_res2)[0]->res;

      /* link changes to continuous level */
      if (! discretemdl)
	addBridgeConditions ((*itms_res2)[0]->already_changed);

      /* update prevTrans in itms_res2. */
      for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator al_it2 = (*itms_res2)[0]->already_changed->begin(); al_it2 != (*itms_res2)[0]->already_changed->end(); al_it2++) {
	if ((*al_it2).second->getLabel() == "false")
	  (*al_it2).second->prevTrans = NULL;
      }
      delete (*itms_res2)[0];
      delete itms_res2;
      itms_res2 = 0;
      delete add_props;
      /* update nextTrans. */
      for (multimap<bool, Atom*, eqbool>::iterator m = curr_trans->pre->atoms->begin(); m != curr_trans->pre->atoms->end(); m++) {
	if ((*current_step->state)[(*m).second->getId()] == NULL) {
	  tra->addComp((*m).second->getVarComp(), current_step, time_step, 1, 0);
#ifdef USE_BDD
	  addToStateBdd(current_step, (*m).second->getVarComp()->getName());
#endif
	}
	(*current_step->state)[(*m).second->getId()]->addToNextTrans(curr_trans);
      }    
    }
    /* add a new transition to be tracked. */
    /*for (multimap<bool, Atom*, eqbool>::iterator cm = curr_trans->pre->atoms->begin(); cm != curr_trans->pre->atoms->end(); cm++)
      if ((*cm).second->getVarType() == "Command")
	(*transitions_CM)[curr_trans->trans_id] = new transitionCM(curr_trans, time_step, (*cm).second);*/
  }
}

/*
   Next State construction.
*/
void Scheduler::constructNextNominalStateLarge(Trajectory* tra, vector<Atom*>* none_coms, vector<Atom*>* coms, bool internal, hash_map<int, transitionCM*, hash<int> >* transitions_CM, int &prev_time, Atom* &curr_com) {
  hash_map<int, Transition*, hash<int> >* al_trans = new hash_map<int, Transition*, hash<int> >();
  vector<ItmsRes*>* itms_res;
  Transition* curr_trans;
  TrajEl* curr_el;
  int time_step = tra->getCurrentTime();
  //std::cerr << "[Debug]:constructNextNominalState: time_step: " << time_step << std::endl;
  State* current_step = tra->getState(time_step);
 
  State* next_step = new State();  /* create a new state */
  tra->addState(time_step + 1, next_step);
#ifdef PROBAS
  (*states_cover)[time_step + 1] = 0;   /* update states cover. */
#endif
  multimap<bool, TrajEl*, eqbool>* add_props;
  
  
  /* add each currently executed multiple commands if it exists. 
   *  ONLY ON EXOGENEOUS COMMANDS !!! to be changed...
   */
  for (hash_map<int, transitionCM*, hash<int> >::iterator curr_coms_it = transitions_CM->begin(); curr_coms_it != transitions_CM->end(); curr_coms_it++) {
    curr_el = getElt((*curr_coms_it).second->trans_trig_cmd->getId(), current_step, tra, time_step, 1);
    lockTrajElsPerm(curr_el, current_step);
    coms->push_back(curr_el->at);    /* add to coms. */
  }

  add_props = new multimap<bool, TrajEl*, eqbool>();
  
  /* padding none commands */
  /*for (unsigned int i=0; i<none_coms->size (); i++) {
    curr_el = getElt((*coms)[i]->getId(), current_step, tra, time_step, 1);
    add_props->insert(pair<bool, TrajEl*>(1, curr_el));
  }
  itms_res = addPropositions(add_props, NULL, current_step, NULL, 1, tra->rank, tra->rank, tra, time_step, 0);
  if (itms_res->empty ())
    cerr << "Error update the current state with none commands: no valid state found.\n";
    current_step = (*itms_res)[0]->res;*/

  add_props = new multimap<bool, TrajEl*, eqbool>(); /* sets of commands at current time. */
  for (unsigned int j=0; j<coms->size(); j++) { /* add others incoming commands. */ 
#ifdef VERB_SCHED
    cerr << "adding : " << (*coms)[j]->getId() << endl;
#endif
    curr_el = getElt((*coms)[j]->getId(), current_step, tra, time_step, 1);
    add_props->insert(pair<bool, TrajEl*>(1, curr_el));
    if (!internal) {   /* lock it. Only if not an 'internal' command. */
      lockTrajElsPerm(curr_el, current_step);
    }
  }
  itms_res = addPropositions(add_props, NULL, current_step, NULL, 1, tra->rank, tra->rank, tra, time_step, 0);
  if (itms_res->empty ())
    cerr << "Error update the current state: no valid state found.\n";
  current_step = (*itms_res)[0]->res;

  if (! discretemdl)  /* link changes to continuous level */
    addBridgeConditions ((*itms_res)[0]->already_changed);
  
  tra->addState(time_step, current_step);
  delete add_props;
  
  for (unsigned int i1 = 0; i1<coms->size(); i1++) {  /* search for directly commanded transition. */
    for (unsigned int k1=0; k1<(*coms)[i1]->ref_trans_pre->size(); k1++) {
      curr_trans = (*(*coms)[i1]->ref_trans_pre)[k1];
      if ((*al_trans)[curr_trans->trans_id] == NULL && curr_trans->getType() == "Nominal" 
	  && !curr_trans->idle) {
	(*al_trans)[curr_trans->trans_id] = curr_trans;
	if (evalStateFormula(curr_trans->pre, tra, time_step)) {
	  triggerTransition(curr_trans, current_step, next_step, tra, transitions_CM, time_step);
	}
      }
    }
  }
  /* search for enabled transitions in changes made to current_step. */
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator i2 = (*itms_res)[0]->already_changed->begin(); 
       i2 != (*itms_res)[0]->already_changed->end(); i2++) {
    for (unsigned int k2=0; k2<(*i2).second->at->ref_trans_pre->size(); k2++) {
      curr_trans = (*(*i2).second->at->ref_trans_pre)[k2];
      if ((*al_trans)[curr_trans->trans_id] == NULL 
	  && curr_trans->getType() == "Nominal" && !curr_trans->idle) {
	(*al_trans)[curr_trans->trans_id] = curr_trans;
	if (evalStateFormula(curr_trans->pre, tra, time_step)) {
	  triggerTransition(curr_trans, current_step, next_step, tra, transitions_CM, time_step);
	}
      }
    }
  }
  delete (*itms_res)[0];
  delete itms_res;
  itms_res = 0;
  /* delete from transitions_CM the transitions that have been completely executed. */
  hash_map<int, transitionCM*, hash<int> >::iterator tCM_it = transitions_CM->begin();
  vector<int>* del_trans = new vector<int>();
  while(tCM_it != transitions_CM->end()) {
    if ((*tCM_it).second->getEnd(time_step)) {
      (*tCM_it).second->trans->showTransition();
      del_trans->push_back((*tCM_it).first);
    }
    tCM_it++;
  }
  for (unsigned int del=0; del<del_trans->size(); del++)
    transitions_CM->erase((*del_trans)[del]);
  
  delete al_trans;
  if (! next_step->state->empty ()) {
    tra->addState(time_step + 1, next_step);
    tra->setTimeCorr (simTime);
  }
  else {
    tra->removeLastState ();
    next_step = 0;
  }
#ifdef VERB_SCHED
  cerr << "constructNextNominalState --> showing traj :\n";
  tra->showTrajectoryTransitions();
  /*for (int i=tra->low_time; i<tra->getCurrentTime (); i++)
    tra->showTrajectory (i);*/
  cerr << "current_step (" << current_step->state->size() << ") :" << endl;
  viewHashMapTrajElLow(current_step->state);
  if (next_step != NULL) {
    cerr << endl << "next_step (" << next_step->state->size() << ") :" << endl;
    viewHashMapTrajElLow(next_step->state);
    cerr << endl;
  }
#endif
}

//running thread of the scheduler on a scenario file.
#ifdef DISCRETE_ONLY
bool Scheduler::runSchD(char* filename) {
  scenario = filename;
  run2();
  return 1;
}

void Scheduler::run2() {
  discretemdl = 1;
  bool internal_com = false;  /* store commands at the previous state to reset the channels to none (default) */
  int search_amp = 0;   /* time amplitude of the search. */
  int traj_time = 0;
 
  set<int, greater<int> >* decision_times = new set<int, greater<int> >();  /* vector for times of decision */
  vector<Atom*>* prev_commands = 0;
  vector<Atom*>* none_commands;
  
  /* variables for transitions CM & internal commands storage. */
  int prev_time = 0; /* time when the current transition has been triggered. */
  Atom* curr_com = 0; /* current command of the current transition. Could be a set? */
  hash_map<int, transitionCM*, hash<int> >* transitions_CM = new hash_map<int, transitionCM*, hash<int> >();
  
  if ((scenar_stream = fopen(scenario, "r")) == NULL) {   /* open the scenario's file. */
    std::cerr << "can't open scenario file: " << scenario << std::endl;
    return;
  }
  scenarEvent* inco_scenar_init;
  scenarEvent* inco_scenar = NULL;
  if ((inco_scenar_init = readFromScenario()) == NULL) {
    cerr << "exiting run2.\n";
    exit (0);  /* TODO: catch the error instead */
  }
  multimap<double, Trajectory*, greater<double> >* cover;
  multimap<double, Trajectory*, greater<double> >* temp_trajs;
  multimap<double, Trajectory*, greater<double> >* obs_res;
  Trajectory* best_traj;
  Trajectory* best_traj2;
  double norm; /* normalizer factor for trajectories probabilities. */
  /*double current_rank = 0;*/ /* current rank of the system : tracked & searched fault number. */
  
  multimap<bool, Atom*, eqbool>* i_commands = NULL; /* incoming commands. For compatibility with new methods (hybrid) */

  (*Eagenda)[1] = inco_scenar_init;   /* initialize the commands handler with inco_scenar. */
  clock_t start, end, end2;
  double elapsed;
  
  if (discrete_time == 0) { 
    for (multimap<double, Trajectory*, greater<double> >::iterator traj_it5 = tracked_traj->begin(); traj_it5 != tracked_traj->end(); traj_it5++) {
      for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator lock_it = (*traj_it5).second->getState(0)->state->begin(); lock_it != (*traj_it5).second->getState(0)->state->end(); lock_it++)
	if ((*lock_it).second->label == "true" && (*lock_it).second->getVarType () != "Command")
	  lockTrajElsPerm((*lock_it).second, (*traj_it5).second->getState(0));
    }
  }

  while((*Eagenda)[(double) (discrete_time+1)] != NULL) {
    cerr << "Time : " << discrete_time+1 << endl;
    cerr << "From scenario:\n";
    inco_scenar = (*Eagenda)[(double)(discrete_time+1)];
    for (unsigned int z=0; z<inco_scenar->incomingEvent->size(); z++) {
      (*inco_scenar->incomingEvent)[z]->showAtom(0);
      cerr << " ";
    }
    cerr << endl;
    /* incoming event is a command. */
    if (inco_scenar->command) {
      discrete_time++;
      start = clock();
      norm = 0;
      delete prev_commands;
      prev_commands = inco_scenar->incomingEvent;
      none_commands = getNoneCommands(prev_commands);
      for (multimap<double, Trajectory*, greater<double> >::iterator traj_it = tracked_traj->begin(); traj_it != tracked_traj->end(); traj_it++) {
	/*if (syncPreviousState ((*traj_it).second))
	  discrete_time++;*/
	/* construct the new step for each trajectory. */
	traj_time = (*traj_it).second->current_time;
	//cerr << "---------> Constructing new state : " << traj_time << endl;
	constructNextNominalStateLarge((*traj_it).second, none_commands, inco_scenar->incomingEvent, internal_com, transitions_CM, prev_time, curr_com);
	  if (traj_time == (*traj_it).second->current_time)  /* no new state has been created */
	    syncPreviousState ((*traj_it).second);
#ifdef PROBAS
	  /* compute probabilities. */
	  if ((*traj_it).second->getState (discrete_time) != NULL) {
	    (*traj_it).second->computeStateProb(discrete_time);
	    norm += (*traj_it).second->getState(discrete_time)->proba;
	  }
#endif      
      }
      /* normalize probabilities. */
#ifdef PROBAS
      for (multimap<double, Trajectory*, greater<double> >::iterator traj_it2 = tracked_traj->begin(); 
	   traj_it2 != tracked_traj->end(); traj_it2++) {
	if ((*traj_it2).second->getState (discrete_time) != NULL) {
	  cerr << "state proba : " << (*traj_it2).second->getState(discrete_time)->proba << endl;
	  (*traj_it2).second->proba = (*traj_it2).second->getState(discrete_time)->proba / norm;
	  cerr << "normalized proba : " << (*traj_it2).second->proba << endl;
	}
      }
#endif
      delete none_commands;
      /* time spent operating. */
      end = clock();
      elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
      cerr << "[Scheduler] --> elapsed time : " << elapsed << endl;
#ifdef BENCHMK
      addBmkEntry(discrete_time, elapsed, 0, tracked_traj->size());
#endif
    }
    else { /* incoming events are observations. */
      start = clock();
      norm = 0;
      traj_time = (*tracked_traj->begin()).second->current_time;
      temp_trajs = new multimap<double, Trajectory*, greater<double> >();
      best_traj = new Trajectory((*tracked_traj->begin()).second, traj_time+1);   /* copy and save the most probable traj in case diagnosis is required. */
      
      /* insert observations into each currently tracked trajectory (rank = 0). */
      for (multimap<double, Trajectory*, greater<double> >::iterator traj_it = tracked_traj->begin(); traj_it != tracked_traj->end(); traj_it++) {
	cerr << "[Scheduler] --> updating tracked trajectories with observations (" << discrete_time << ")\n";
	i_commands = new multimap<bool, Atom*, eqbool> ();
	traj_time = (*traj_it).second->current_time;
	for (unsigned int i=0; i<inco_scenar->incomingEvent->size (); i++)
	  i_commands->insert (pair<bool, Atom*> (1, (*inco_scenar->incomingEvent)[i]));
	
	obs_res = observationResults(i_commands, (*traj_it).second, traj_time, 0, 1);
	if ((obs_res->size() == 1) && getConflicts((*obs_res->begin()).second, traj_time, 1)->empty()) {
	  end2 = clock();
	  elapsed = ((double)(end2 - start))/CLOCKS_PER_SEC;
	  cerr << "[Scheduler] --> elapsed time (observations insertion) : " << elapsed << endl;
	  temp_trajs->insert(pair<double, Trajectory*>((*obs_res->begin()).first, (*obs_res->begin()).second));
	  norm += (*obs_res->begin()).first;
	  tracked_traj->erase(traj_it);
	}
	else {
	  end2 = clock();
	  elapsed = ((double)(end2 - start))/CLOCKS_PER_SEC;
	  cerr << "[Scheduler] --> elapsed time (observations insertion) : " << elapsed << endl;
	  tracked_traj->erase(traj_it);
	}
      }
      
      /* in case observations were consistent with at least one tracked trajectory. */
      if (!temp_trajs->empty()) {
	for (multimap<double, Trajectory*, greater<double> >::iterator temp_it = temp_trajs->begin(); temp_it != temp_trajs->end(); temp_it++) {
	  
#ifdef PROBAS
	   /* normalize proba as traj is inserted .*/
	  tracked_traj->insert(pair<double, Trajectory*>((*temp_it).first / norm, (*temp_it).second));
	  (*temp_it).second->proba = (*temp_it).first / norm;
	  cerr << "normalized proba after inserting obs : " << (*temp_it).first / norm << endl;
#else
	  tracked_traj->insert(pair<double, Trajectory*>((*temp_it).second->rank, (*temp_it).second));
#endif
	  
	}
	end = clock(); 	/* time spent operating. */
	elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
	cerr << "[Scheduler] --> elapsed time : " << elapsed << endl;
#ifdef BENCHMK
	addBmkEntry(discrete_time, elapsed, 0, tracked_traj->size());
#endif
	delete best_traj;
      }
      else {   /* in case no more trajectory is consistent with new observations, perform a full diagnosis. */
	cerr << "---------> Diagnosing...\n";
	traj_time = best_traj->current_time;
	best_traj2 = new Trajectory(best_traj, traj_time + 1);  	/* copy and save a trajectory for reconfiguration. */

	cover = coverTrack3(best_traj, current_rank + 1, best_traj->rank + 10, i_commands, search_amp, current_rank);
	/*cover = coverTrack2(best_traj, current_rank+1, best_traj->rank + 10, i_commands, search_amp, current_rank);*/

	//graphTraj(cover);
	delete tracked_traj;
	tracked_traj = cover;
	end = clock(); 	/* time spent operating. */
	elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
	cerr << "[Scheduler] --> elapsed time : " << elapsed << endl;
#ifdef BENCHMK
	addBmkEntry(discrete_time, elapsed, 1, tracked_traj->size());
#endif
#ifdef RECONF
	/* Should better wait a bit before reconfiguring. */
	/*vector<Trajectory*>* snom = new vector<Trajectory*>();
	  vector<Trajectory*>* sfail = new vector<Trajectory*>();
	  snom->push_back(new Trajectory(best_traj2, best_traj2->getCurrentTime () + 1));
	  sfail->push_back(new Trajectory((*tracked_traj->begin()).second, (*tracked_traj->begin()).second->getCurrentTime () + 1));
	  searchMinReconf(snom, sfail);*/
#endif
	delete best_traj2;
      }
      decision_times->insert(traj_time);   /* add the current time as a decision time. */
      delete temp_trajs;
      discrete_time++;
    } /* end of scenario */
    Eagenda->erase((double)discrete_time);
    /*internal_com = organizeCommands(discrete_time, inco_scenar->command);*/  /* rank incoming commands */
    internal_com = false;
    (*Eagenda)[(double)(discrete_time+1)] = readFromScenario ();

    /*if (discrete_time == 0) {*/  /* locking the first state of the nominal traj. */
    /*for (multimap<double, Trajectory*, greater<double> >::iterator traj_it5 = tracked_traj->begin(); traj_it5 != tracked_traj->end(); traj_it5++) {
	for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator lock_it = (*traj_it5).second->getState(0)->state->begin(); lock_it != (*traj_it5).second->getState(0)->state->end(); lock_it++)
	  if ((*lock_it).second->label == "true")
	    lockTrajElsPerm((*lock_it).second, (*traj_it5).second->getState(0));
      }
      }*/
    /*if (inco_scenar != NULL)
     if (inco_scenar->command)
     discrete_time++;*/
   inco_scenar = 0;
  }
  /*graphTraj(tracked_traj);*/
  delete transitions_CM;
  delete inco_scenar;
}
#endif /* DISCRETE_ONLY */

void Scheduler::addBridgeConditions (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = alchg->begin(); it != alchg->end(); it++) {
    if ((*it).second->at->cond) {
      if ((*it).second->getLabel () == "true")
	addBridgeCond ((*it).second->at, 1, current_rt);
      else addBridgeCond ((*it).second->at, 0, current_rt);
    }
  }
}

void Scheduler::addContEvents (Trajectory* tr) {
  /*syncPreviousState (tr);*/
  int dt = tr->getCurrentTime ();
  /*State* current_state = tr->getLastState ();*/
  State* current_state = tr->getState (dt);
  multimap<bool, TrajEl*, eqbool>* addProps = new multimap<bool, TrajEl*, eqbool>();
  vector<ItmsRes*>* itms_res;
  TrajEl* elt;

  for (multimap<bool, Atom*, eqbool>::iterator it = contEvents->begin(); 
       it != contEvents->end(); it++) {
    if ((*it).first) { /* true */
      elt = getElt((*it).second->getId(), current_state, tr, dt);
      lockTrajElsPerm(elt, current_state);
    }
    else {
      elt = getElt((*it).second->getId(), current_state, tr, dt);
      /*elt->setLabel ("false");*/
      elt->lock = 4;
    }
    addProps->insert(pair<bool, TrajEl*>((*it).first, elt));
  }
  
  itms_res = addPropositions(addProps, NULL, current_state, NULL, 1, tr->rank, tr->rank, tr, dt, 0);
  
  if (itms_res->size() < 1) {
    cerr << "Error adding continuous events: no valid state found.\n";
    exit (1);
  }

  /* look for commands */
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = (*itms_res)[0]->already_changed->begin (); it != (*itms_res)[0]->already_changed->end (); it++) {
    /*cerr << "already changed: " << (*it).second->getId () << " -- " << (*it).second->at->ref_trans_pre->empty () << endl;*/
    /*if ((*it).second->getVarType () == "Command" && (*it).second->getLabel () == "true") { */ /* qualitative commands */
    /*cerr << "at: " << (*it).second->getId () << " -- " << (*it).second->at->ref_trans_pre->empty () << endl;*/
    if (! (*it).second->at->ref_trans_pre->empty () && (*it).second->getLabel () == "true") {
#ifdef VERB_SCHED
      cerr << "command detected: " << (*it).second->getId () << endl;
#endif
      comEvents->push_back ((*it).second->at);
    }
  }
#ifdef VERB_SCHED    
  cerr << "current_state (" << (*itms_res)[0]->res->state->size() << ") :" << endl;
  cerr << (*itms_res)[0]->res->comps->size () << endl;
  viewHashMapTrajElLow((*itms_res)[0]->res->state);
  cerr << endl;
#endif
  tr->addState (dt, (*itms_res)[0]->res);
  contEvents->clear();
}

bool Scheduler::syncPreviousState (Trajectory* tr) {
  if (!tr->getLastState ()->state->empty () 
      && tr->getLastState ()->st_lock == true) {
#ifdef VERB_SCHED    
    cerr << "Scheduler --> syncing state (" << tr->getLastState ()->state->size () << ")\n";
    viewHashMapTrajElLow(tr->getLastState ()->state);
    cerr << endl;
#endif
    tr->addState (tr->getCurrentTime () + 1, new State());
    tr->setTimeCorr (simTime);
    return true;
  }
  return false;
}

unsigned int Scheduler::doCycle (Trajectory* tr, runtime rt, unsigned int mode, double stime) {
  scenarEvent* inco_scenar = 0;
  vector<Atom*>* prev_commands = 0;
  vector<Atom*>* none_commands;
  bool internal_com = false;
  bool com = false;
  bool ce = false;        /* comEvent */
  hash_map<int, transitionCM*, hash<int> >* transitions_CM = new hash_map<int, transitionCM*, hash<int> >();
  int prev_time = 0;
  Atom* curr_com = 0;

  simTime = stime;   /* static link to continuous time line */
  current_rt = rt;
  if (! contEvents->empty()) {    /* incoming events from continuous level. */
    syncPreviousState (tr);
    addContEvents (tr);
  }

  if ((! Eagenda->empty () && Eagenda->count (stime) != 0)
      || (ce = !comEvents->empty())) {
    if (! Eagenda->empty () && Eagenda->count (stime) != 0)
      inco_scenar = (*Eagenda)[stime];
      /*|| ! isNull (rt->bridgeConflicts)) {*/
    if (inco_scenar == NULL) { /* && isNull (rt->bridgeConflicts)) { */
      com = true;
      prev_commands = new vector<Atom*>(*comEvents);
    }
    else { /*if (! isNull (inco_scenar)) {*/
      com = inco_scenar->command;
      if (com && ce) {
	prev_commands = inco_scenar->incomingEvent;
	for (unsigned int i=0; i<comEvents->size(); i++)
	  prev_commands->push_back((*comEvents)[i]);
      }
      else if (com)
	prev_commands = inco_scenar->incomingEvent;
    }
    /*else {*/ /* conflicts from diag at continuous level */
      /*com = false;
    }*/
    
    if (com) {                 /* incoming events are commands */
      /*none_commands = getNoneCommands(prev_commands);*/ /* DEPRECATED */
      syncPreviousState (tr);  /* must synchronize in case of a previously locked state. */
#ifdef VERB_SCHED
      cerr << "---------> Constructing new state : " << tr->getCurrentTime () + 1 << endl;
#endif
      for (unsigned int i=0; i<prev_commands->size (); i++)
	cerr << (*prev_commands)[i]->getId () << endl;
      constructNextNominalStateLarge(tr, none_commands, prev_commands, internal_com, transitions_CM, prev_time, curr_com); 
      comEvents->clear();
      /*delete none_commands;*/
      delete prev_commands;
    }
    /*else {        */                              /* incoming events are observations or conflicts */
      /* TODO */
      /*multimap<double, Trajectory*, greater<double> >* obs_res;
      Diagnose ();*/
    /*}*/
  }
  setRTDTimeStep (rt, tr->getCurrentTime ());   /* update runtime discrete time ref */
  delete transitions_CM;
  return 1;
}

void Scheduler::Diagnose (Trajectory* tr, runtime rt, multimap<bool, Atom*, eqbool>* obs, multimap<double, hybTrajectory*, greater<double> >* tt) {
  clock_t start, end;
  double norm, elapsed;
  int search_amp = 0; /* time amplitude of the search */
  multimap<double, Trajectory*, greater<double> >* obs_res;
  multimap<double, Trajectory*, greater<double> >* temp_trajs = new multimap<double, Trajectory*, greater<double> >();
  current_rt = rt;

  start = clock ();
  norm = 0;
  current_rank = tr->rank;
  
  /*cerr << "observations: ";
    for (multimap<bool, Atom*, eqbool>::iterator it = obs->begin (); it != obs->end (); it++)
    cerr << (*it).second->getId () << " (" << (*it).first << ") -- ";
    cerr << endl;*/

  Trajectory* ctr = new Trajectory (tr, tr->getCurrentTime () + 1);  /* copy the trajectory for later use */
  cerr << "[Scheduler] --> updating tracked trajectories with observations (" << tr->getCurrentTime () << ")\n";
  obs_res = observationResults(obs, tr, tr->getCurrentTime (), 0, 1);        /* insert observations into each currently tracked trajectory (rank = 0) */
  
  if ((obs_res->size() == 1) 
      && getConflicts((*obs_res->begin()).second, (*obs_res->begin ()).second->getCurrentTime (), 1)->empty()) {    /* in case it is just a glitche... */
    end = clock();
    elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
    cerr << "[Scheduler] --> elapsed time (observations insertion) : " << elapsed << endl;
    temp_trajs->insert(pair<double, Trajectory*>((*obs_res->begin()).first, (*obs_res->begin()).second));
    norm += (*obs_res->begin()).first;
  }
  else {
    end = clock();
    elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
    cerr << "[Scheduler] --> elapsed time (observations insertion) : " << elapsed << endl;
  }
  
  delete obs_res;
  
  if (!temp_trajs->empty()) { /* in case observations were consistent with at least one tracked trajectory. */
#ifdef USE_BDD
    for (multimap<double, Trajectory*, greater<double> >::iterator temp_it = temp_trajs->begin(); temp_it != temp_trajs->end(); temp_it++) {
      (*temp_it).first = (*temp_it).first / norm;       /* normalize proba as traj is inserted .*/
      (*temp_it).second->proba = (*temp_it).first / norm;
    }
#endif
  }
  else {  /* in case no more trajectory is consistent with new observations, perform a full diagnosis. */
    delete temp_trajs;
    temp_trajs = 0;
    cerr << "---------> Diagnosing...\n";
    /*cerr << "starting traj: " << ctr << endl;
      ctr->showTrajectoryTransitions ();*/
    temp_trajs = coverTrack3(ctr, current_rank + 1, ctr->rank + 3, obs, search_amp, current_rank);
    end = clock();
    elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
    cerr << "[Scheduler] --> elapsed time : " << elapsed << endl;
  }
  for (multimap<double, Trajectory*, greater<double> >::iterator it = temp_trajs->begin (); it != temp_trajs->end (); it++) {
    hybTrajectory* nht = new hybTrajectory ();
    nht->traj = (*it).second;
    nht->rt = ce_copyRuntime (rt);
    tt->insert (pair<double, hybTrajectory*>((*it).first, nht));
  }
  delete temp_trajs;
}

void Scheduler::raz () {
  discrete_time = 0;
  tracked_traj->clear();
  delete Eagenda;
  Eagenda = new map<double, scenarEvent*>();
}

/* graph trajectories. */
void Scheduler::graphTraj(multimap<double, Trajectory*, greater<double> >* cov) {
  string file_name = "traj";
  char* s;
  char buf[20];
  s = buf;
  int i = 0;
  for (multimap<double, Trajectory*, greater<double> >::iterator it = cov->begin(); it != cov->end(); it++) {
    sprintf(buf, "%i", i);
    toDot((*it).second, string(file_name + s));
    i++;  
  }
}
