/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef SCHEDULER_H
#define SCHEDULER_H

#include <sys/types.h>
#include <unistd.h>
#include <stddef.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>
#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map>
#endif
#include <algorithm>
#include <queue>
#include <map>
#include "Diagnoser.h"
#include "hybTrajectory.h"
#include "model_tools.h"

#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif

using namespace std;

/**
* \class scenarEvent
* \brief Scenario event for testing the system.
*/

class scenarEvent {
 public:
  bool command;
  vector<Atom*>* incomingEvent;
  scenarEvent(bool com, vector<Atom*>* event) {
    command = com;
    incomingEvent = event;
  }
  scenarEvent() {}
  ~scenarEvent() {
    delete incomingEvent;
  }
};

/**
* \class transitionCM
* \brief Transition for Multiple Commands, ie. chains of commands resulting of a transition.
*/

class transitionCM {
 public:
  Transition* trans;
  int trans_trig_time;
  /* command that triggered the trans. This command has a duration of trans->time_length. */
  Atom* trans_trig_cmd;
  transitionCM(Transition* tr, int ttt, Atom* at) {
    trans = tr;
    trans_trig_time = ttt;
    trans_trig_cmd = at;
  }
  ~transitionCM(){
  }
  /* evaluate if the transition is terminated. */
  bool getEnd(int time_step) {
    return ((trans->time_length + trans_trig_time) == time_step);
  }
};

extern void addBridgeCond (Atom* at, bool b, runtime rt);
extern void addBridgeTrans (Transition* tr, runtime rt);
extern char scenarFile[40];

/**
 * \class Scheduler
 * \brief Scheduler for handling commands, processing prevision states and driving the diagnosis.
 * 
 */

class Scheduler : public Diagnoser {

  struct timespec delay;
  bool workFlag;

  static double simTime;  /* simulation time at continuous level */

 public:
  int current_command_num;
  int discrete_time;                  /**< number of discrete time iterations of the system. */
  char* scenario;                            /**< filename for scenario. */
  FILE* scenar_stream;                       /**< file for scenario. */
  map<double, scenarEvent*>* Eagenda;           /**< agenda of incoming events. */
  multimap<bool, Atom*, eqbool>* contEvents; /**< incoming continuous events. */
  vector<Atom*>* comEvents;                  /**< incoming continuous commands. */
  runtime current_rt;
  double current_rank;

  Scheduler(string mn);
  ~Scheduler();
  bool compile();
  int initSchD();
  void setRunningStatus(bool st);
  void addCommandEvent(Atom* com, int start_time);
  bool organizeCommands(int time_step, bool increase);
  vector<Atom*>* getNoneCommands(vector<Atom*>* coms); //return the none command corresponding to the commands as arguments.
  scenarEvent* readFromScenario();
  void loadScenario();  /* load a scenario with dated events */
  void triggerTransition(Transition* curr_trans, State* &current_step, State* &next_step, Trajectory* tra, hash_map<int, transitionCM*, hash<int> >* transitions_CM, int time_step);
  void constructNextNominalStateLarge(Trajectory* tra, vector<Atom*>* none_coms, vector<Atom*>* coms, bool internal, hash_map<int, transitionCM*, hash<int> >* transitionsCM, int &prev_time, Atom* &curr_com);
  
  bool runSchD(char* filename);
  void run();
  void run2();

  /* bridging with caen (koala) */ 
  void addBridgeConditions (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg);
  void addContEvents (Trajectory* tr);
  bool syncPreviousState (Trajectory* tr);
  unsigned int doCycle (Trajectory* tr, runtime rt, unsigned int mode, double stime);
  void Diagnose (Trajectory* tr, runtime rt, multimap<bool, Atom*, eqbool>* obs, multimap<double, hybTrajectory*, greater<double> >* tt);
  void graphTraj(multimap<double, Trajectory*, greater<double> >* cov);
  void raz();
};
#endif
